@@ -12,7 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import os, pickle
+import os, pickle, re
 import build
 import dependencies
 import mesonlib
@@ -37,10 +37,9 @@ class TestSerialisation:
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
 class Backend():
-    def __init__(self, build, interp):
+    def __init__(self, build):
         self.build = build
         self.environment = build.environment
-        self.interpreter = interp
         self.processed_targets = {}
         self.dep_rules = {}
         self.build_to_src = os.path.relpath(self.environment.get_source_dir(),
@@ -69,12 +68,18 @@ class Backend():
         return filename
 
     def get_target_dir(self, target):
-        dirname = target.get_subdir()
+        if self.environment.coredata.get_builtin_option('layout') == 'mirror':
+            dirname = target.get_subdir()
+        else:
+            dirname = 'meson-out'
         os.makedirs(os.path.join(self.environment.get_build_dir(), dirname), exist_ok=True)
         return dirname
 
     def get_target_private_dir(self, target):
-        dirname = os.path.join(self.get_target_dir(target), target.get_basename() + target.type_suffix())
+        return os.path.join(self.get_target_dir(target), target.get_basename() + target.type_suffix())
+
+    def get_target_private_dir_abs(self, target):
+        dirname = self.get_target_private_dir(target)
         os.makedirs(os.path.join(self.environment.get_build_dir(), dirname), exist_ok=True)
         return dirname
 
@@ -87,7 +92,7 @@ class Backend():
             language = comp.get_language()
             suffix = '.' + comp.get_default_suffix()
             if language not in langlist:
-                outfilename = os.path.join(self.get_target_private_dir(target), target.name + '-unity' + suffix)
+                outfilename = os.path.join(self.get_target_private_dir_abs(target), target.name + '-unity' + suffix)
                 outfileabs = os.path.join(self.environment.get_build_dir(), outfilename)
                 outfileabs_tmp = outfileabs + '.tmp'
                 abs_files.append(outfileabs)
@@ -167,7 +172,7 @@ class Backend():
 
     def determine_ext_objs(self, extobj, proj_dir_to_build_root=''):
         result = []
-        targetdir = self.get_target_private_dir(extobj.target)
+        targetdir = self.get_target_private_dir_abs(extobj.target)
         suffix = '.' + self.environment.get_object_suffix()
         for osrc in extobj.srclist:
             osrc_base = osrc.fname
@@ -187,7 +192,7 @@ class Backend():
 
     def get_pch_include_args(self, compiler, target):
         args = []
-        pchpath = self.get_target_private_dir(target)
+        pchpath = self.get_target_private_dir_abs(target)
         includeargs = compiler.get_include_args(pchpath)
         for lang in ['c', 'cpp']:
             p = target.get_pch(lang)
@@ -203,15 +208,16 @@ class Backend():
     def generate_basic_compiler_args(self, target, compiler):
         commands = []
         commands += compiler.get_always_args()
-        if self.environment.coredata.buildtype != 'plain':
-            commands += compiler.get_warn_args(self.environment.coredata.warning_level)
+        if self.environment.coredata.get_builtin_option('buildtype') != 'plain':
+            commands += compiler.get_warn_args(self.environment.coredata.get_builtin_option('warning_level'))
+        commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
         commands += self.build.get_global_args(compiler)
         commands += self.environment.coredata.external_args[compiler.get_language()]
         commands += target.get_extra_args(compiler.get_language())
-        commands += compiler.get_buildtype_args(self.environment.coredata.buildtype)
-        if self.environment.coredata.coverage:
+        commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        if self.environment.coredata.get_builtin_option('coverage'):
             commands += compiler.get_coverage_args()
-        if self.environment.coredata.werror:
+        if self.environment.coredata.get_builtin_option('werror'):
             commands += compiler.get_werror_args()
         if isinstance(target, build.SharedLibrary):
             commands += compiler.get_pic_args()
@@ -220,10 +226,10 @@ class Backend():
             if isinstance(target, build.Executable):
                 commands += dep.get_exe_args()
 
-        # Fortran rquires extra include directives.
+        # Fortran requires extra include directives.
         if compiler.language == 'fortran':
             for lt in target.link_targets:
-                priv_dir = os.path.join(lt.subdir, lt.get_basename() + lt.type_suffix())
+                priv_dir = os.path.join(self.get_target_dir(lt), lt.get_basename() + lt.type_suffix())
                 incflag = compiler.get_include_args(priv_dir)
                 commands += incflag
         return commands
@@ -252,10 +258,16 @@ class Backend():
         links to and return them so they can be used in unit
         tests.'''
         if not isinstance(target, build.Executable):
-            print(target)
             return []
-        prospectives = target.get_transitive_rpaths()
-        return [os.path.join(self.environment.get_build_dir(), i) for i in prospectives if len(i) > 0]
+        prospectives = target.get_transitive_link_deps()
+        result = []
+        for ld in prospectives:
+            if ld == '' or ld == '.':
+                continue
+            dirseg = os.path.join(self.environment.get_build_dir(), self.get_target_dir(ld))
+            if dirseg not in result:
+                result.append(dirseg)
+        return result
 
     def write_test_file(self, datafile):
         arr = []
@@ -290,9 +302,10 @@ class Backend():
             outdir = self.environment.scratch_dir
             fname = os.path.join(outdir, p.filebase + '.pc')
             ofile = open(fname, 'w')
-            ofile.write('prefix=%s\n' % self.environment.get_coredata().prefix)
-            ofile.write('libdir=${prefix}/%s\n' % self.environment.get_coredata().libdir)
-            ofile.write('includedir=${prefix}/%s\n\n' % self.environment.get_coredata().includedir)
+            coredata = self.environment.get_coredata()
+            ofile.write('prefix=%s\n' % coredata.get_builtin_option('prefix'))
+            ofile.write('libdir=${prefix}/%s\n' % coredata.get_builtin_option('libdir'))
+            ofile.write('includedir=${prefix}/%s\n\n' % coredata.get_builtin_option('includedir'))
             ofile.write('Name: %s\n' % p.name)
             if len(p.description) > 0:
                 ofile.write('Description: %s\n' % p.description)
@@ -321,3 +334,84 @@ class Backend():
         mfobj['projects'] = self.build.dep_manifest
         open(ifilename, 'w').write(json.dumps(mfobj))
         d.data.append([ifilename, ofilename])
+
+    def get_regen_filelist(self):
+        '''List of all files whose alteration means that the build
+        definition needs to be regenerated.'''
+        deps = [os.path.join(self.build_to_src, df) \
+                for df in self.interpreter.get_build_def_files()]
+        if self.environment.is_cross_build():
+            deps.append(os.path.join(self.build_to_src,
+                                     self.environment.coredata.cross_file))
+        deps.append('meson-private/coredata.dat')
+        if os.path.exists(os.path.join(self.environment.get_source_dir(), 'meson_options.txt')):
+            deps.append(os.path.join(self.build_to_src, 'meson_options.txt'))
+        for sp in self.build.subprojects.keys():
+            fname = os.path.join(self.environment.get_source_dir(), sp, 'meson_options.txt')
+            if os.path.isfile(fname):
+                deps.append(os.path.join(self.build_to_src, sp, 'meson_options.txt'))
+        return deps
+
+    def exe_object_to_cmd_array(self, exe):
+        if self.environment.is_cross_build() and \
+                isinstance(exe, build.BuildTarget) and exe.is_cross:
+            if 'exe_wrapper'  not in self.environment.cross_info:
+                s = 'Can not use target %s as a generator because it is cross-built\n'
+                s += 'and no exe wrapper is defined. You might want to set it to native instead.'
+                s = s % exe.name
+                raise MesonException(s)
+        if isinstance(exe, build.BuildTarget):
+            exe_arr = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))]
+        else:
+            exe_arr = exe.get_command()
+        return exe_arr
+
+    def eval_custom_target_command(self, target, absolute_paths=False):
+        ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
+        srcs = []
+        outdir = self.get_target_dir(target)
+        if absolute_paths:
+            outdir = os.path.join(self.environment.get_build_dir(), outdir)
+        for i in target.sources:
+            if isinstance(i, str):
+                fname = os.path.join(self.build_to_src, target.subdir, i)
+            else:
+                fname = i.rel_to_builddir(self.build_to_src)
+            if absolute_paths:
+                fname = os.path.join(self.environment.get_build_dir(), fname)
+            srcs.append(fname)
+        cmd = []
+        for i in target.command:
+            if isinstance(i, build.Executable):
+                cmd += self.exe_object_to_cmd_array(i)
+                continue
+            if isinstance(i, build.CustomTarget):
+                # GIR scanner will attempt to execute this binary but
+                # it assumes that it is in path, so always give it a full path.
+                tmp = i.get_filename()[0]
+                i = os.path.join(self.get_target_dir(i), tmp)
+            for (j, src) in enumerate(srcs):
+                i = i.replace('@INPUT%d@' % j, src)
+            for (j, res) in enumerate(ofilenames):
+                i = i.replace('@OUTPUT%d@' % j, res)
+            if i == '@INPUT@':
+                cmd += srcs
+            elif i == '@OUTPUT@':
+                cmd += ofilenames
+            else:
+                if '@OUTDIR@' in i:
+                    i = i.replace('@OUTDIR@', outdir)
+                elif '@PRIVATE_OUTDIR_' in i:
+                    match = re.search('@PRIVATE_OUTDIR_(ABS_)?([-a-zA-Z0-9.@:]*)@', i)
+                    source = match.group(0)
+                    if match.group(1) is None and not absolute_paths:
+                        lead_dir = ''
+                    else:
+                        lead_dir = self.environment.get_build_dir()
+                    target_id = match.group(2)
+                    i = i.replace(source,
+                                  os.path.join(lead_dir,
+                                               outdir))
+                cmd.append(i)
+        cmd = [i.replace('\\', '/') for i in cmd]
+        return (srcs, ofilenames, cmd)
