@@ -124,11 +124,15 @@ class Build:
         return self.global_args.get(compiler.get_language(), [])
 
 class IncludeDirs():
-    def __init__(self, curdir, dirs):
+    def __init__(self, curdir, dirs, extra_build_dirs=None):
         self.curdir = curdir
         self.incdirs = dirs
         # Interpreter has validated that all given directories
         # actually exist.
+        if extra_build_dirs is None:
+            self.extra_build_dirs = []
+        else:
+            self.extra_build_dirs = extra_build_dirs
 
     def get_curdir(self):
         return self.curdir
@@ -136,6 +140,9 @@ class IncludeDirs():
     def get_incdirs(self):
         return self.incdirs
 
+    def get_extra_build_dirs(self):
+        return self.extra_build_dirs
+
 class ExtractedObjects():
     def __init__(self, target, srclist):
         self.target = target
@@ -170,7 +177,13 @@ class BuildTarget():
         self.validate_sources()
 
     def get_id(self):
-        return self.subproject + ':' + self.name + self.type_suffix()
+        # This ID must also be a valid file name on all OSs.
+        # It should also avoid shell metacharacters for obvious
+        # reasons.
+        base = self.name + self.type_suffix()
+        if self.subproject == '':
+            return base
+        return self.subproject + '@@' + base
 
     def check_unknown_kwargs(self, kwargs):
         # Override this method in derived classes that have more
@@ -267,13 +280,13 @@ class BuildTarget():
     def extract_all_objects(self):
         return ExtractedObjects(self, self.sources)
 
-    def get_rpaths(self):
-        return self.get_transitive_rpaths()
+    def get_all_link_deps(self):
+        return self.get_transitive_link_deps()
 
-    def get_transitive_rpaths(self):
+    def get_transitive_link_deps(self):
         result = []
         for i in self.link_targets:
-            result += i.get_rpaths()
+            result += i.get_all_link_deps()
         return result
 
     def get_custom_install_dir(self):
@@ -422,6 +435,8 @@ class BuildTarget():
         return self.include_dirs
 
     def add_external_deps(self, deps):
+        if not isinstance(deps, list):
+            deps = [deps]
         for dep in deps:
             if hasattr(dep, 'held_object'):
                 dep = dep.held_object
@@ -430,6 +445,7 @@ class BuildTarget():
                 self.add_include_dirs(dep.include_directories)
                 for l in dep.libraries:
                     self.link(l)
+                self.add_external_deps(dep.ext_deps)
             elif isinstance(dep, dependencies.Dependency):
                 self.external_deps.append(dep)
                 self.process_sourcelist(dep.get_sources())
@@ -448,6 +464,8 @@ class BuildTarget():
             if not isinstance(t, StaticLibrary) and \
             not isinstance(t, SharedLibrary):
                 raise InvalidArguments('Link target is not library.')
+            if self.is_cross != t.is_cross:
+                raise InvalidArguments('Tried to mix cross built and native libraries in target %s.' % self.name)
             self.link_targets.append(t)
 
     def set_generated(self, genlist):
@@ -539,7 +557,7 @@ class Generator():
             if not isinstance(rule, str):
                 raise InvalidArguments('"output" may only contain strings.')
             if not '@BASENAME@' in rule and not '@PLAINNAME@' in rule:
-                raise InvalidArguments('"outputs" must contain @BASENAME@ or @PLAINNAME@.')
+                raise InvalidArguments('Every element of "output" must contain @BASENAME@ or @PLAINNAME@.')
             if '/' in rule or '\\' in rule:
                 raise InvalidArguments('"outputs" must not contain a directory separator.')
         if len(outputs) > 1:
@@ -607,6 +625,8 @@ class StaticLibrary(BuildTarget):
             raise InvalidArguments('Static libraries not supported for C#.')
         self.prefix = environment.get_static_lib_prefix()
         self.suffix = environment.get_static_lib_suffix()
+        if len(self.sources) > 0 and self.sources[0].endswith('.rs'):
+            self.suffix = 'rlib'
         self.filename = self.prefix + self.name + '.' + self.suffix
 
     def get_import_filename(self):
@@ -629,6 +649,8 @@ class SharedLibrary(BuildTarget):
         else:
             self.prefix = environment.get_shared_lib_prefix()
             self.suffix = environment.get_shared_lib_suffix()
+        if len(self.sources) > 0 and self.sources[0].endswith('.rs'):
+            self.suffix = 'rlib'
         self.importsuffix = environment.get_import_lib_suffix()
         self.filename = self.prefix + self.name + '.' + self.suffix
 
@@ -648,8 +670,8 @@ class SharedLibrary(BuildTarget):
     def get_import_filename(self):
         return self.prefix + self.name + '.' + self.importsuffix
 
-    def get_rpaths(self):
-        return [self.subdir] + self.get_transitive_rpaths()
+    def get_all_link_deps(self):
+        return [self] + self.get_transitive_link_deps()
 
     def get_filename(self):
         '''Works on all platforms except OSX, which does its own thing.'''
@@ -694,13 +716,16 @@ class CustomTarget:
                     'install' : True,
                     'install_dir' : True,
                     'build_always' : True,
-                    'depends' : True}
+                    'depends' : True,
+                    'depend_files' : True,
+                    }
 
     def __init__(self, name, subdir, kwargs):
         self.name = name
         self.subdir = subdir
         self.dependencies = []
         self.extra_depends = []
+        self.depend_files = [] # Files that this target depends on but are not on the command line.
         self.process_kwargs(kwargs)
         self.extra_files = []
         self.install_rpath = ''
@@ -744,13 +769,7 @@ class CustomTarget:
                 final_cmd += c.get_command()
             elif isinstance(c, BuildTarget) or isinstance(c, CustomTarget):
                 self.dependencies.append(c)
-                # GIR scanner will attempt to execute this binary but
-                # it assumes that it is in path, so always give it a full path.
-                tmp = c.get_filename()
-                if isinstance(tmp, str):
-                    tmp =[tmp]
-                totarget = [os.path.join('.', c.get_subdir(), i) for i in tmp]
-                final_cmd += totarget
+                final_cmd.append(c)
             elif isinstance(c, list):
                 # Hackety hack, only supports one level of flattening. Should really
                 # work to arbtrary depth.
@@ -784,6 +803,15 @@ class CustomTarget:
             if not isinstance(ed, CustomTarget) and not isinstance(ed, BuildTarget):
                 raise InvalidArguments('Can only depend on toplevel targets.')
             self.extra_depends.append(ed)
+        depend_files = kwargs.get('depend_files', [])
+        if not isinstance(depend_files, list):
+            depend_files = [depend_files]
+        for i in depend_files:
+            if isinstance(i, (File, str)):
+                self.depend_files.append(i)
+            else:
+                mlog.debug(i)
+                raise InvalidArguments('Unknown type in depend_files.')
 
     def get_basename(self):
         return self.name
