@@ -17,6 +17,7 @@ import tempfile
 import mesonlib
 import mlog
 from coredata import MesonException
+import coredata
 
 """This file contains the data files of all compilers Meson knows
 about. To support a new compiler, add its information below.
@@ -52,7 +53,7 @@ gnulike_buildtype_args = {'plain' : [],
                           'release' : ['-O3']}
 
 msvc_buildtype_args = {'plain' : [],
-                       'debug' : ["/MDd", "/Zi", "/Ob0", "/Od", "/RTC1"],
+                       'debug' : ["/MDd", "/ZI", "/Ob0", "/Od", "/RTC1"],
                        'debugoptimized' : ["/MD", "/Zi", "/O2", "/Ob1", "/D"],
                        'release' : ["/MD", "/O2", "/Ob2"]}
 
@@ -91,6 +92,13 @@ mono_buildtype_args = {'plain' : [],
                        'debugoptimized': ['-debug', '-optimize+'],
                        'release' : ['-optimize+']}
 
+gnu_winlibs = ['-lkernel32', '-luser32', '-lgdi32', '-lwinspool', '-lshell32',
+               '-lole32', '-loleaut32', '-luuid', '-lcomdlg32', '-ladvapi32']
+
+msvc_winlibs = ['kernel32.lib', 'user32.lib', 'gdi32.lib',
+                'winspool.lib', 'shell32.lib', 'ole32.lib', 'oleaut32.lib',
+                'uuid.lib', 'comdlg32.lib', 'advapi32.lib']
+
 def build_unix_rpath_args(build_dir, rpath_paths, install_rpath):
         if len(rpath_paths) == 0 and len(install_rpath) == 0:
             return []
@@ -118,15 +126,55 @@ class RunResult():
         self.stdout = stdout
         self.stderr = stderr
 
-class CCompiler():
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+class Compiler():
+    def __init__(self, exelist, version):
         if type(exelist) == type(''):
             self.exelist = [exelist]
         elif type(exelist) == type([]):
             self.exelist = exelist
         else:
-            raise TypeError('Unknown argument to CCompiler')
+            raise TypeError('Unknown argument to Compiler')
         self.version = version
+
+    def get_always_args(self):
+        return []
+
+    def get_linker_always_args(self):
+        return []
+
+    def get_options(self):
+        return {} # build afresh every time
+
+    def get_option_compile_args(self, options):
+        return []
+
+    def get_option_link_args(self, options):
+        return []
+
+    def has_header(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support header checks.' % self.language)
+
+    def compiles(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support compile checks.' % self.language)
+
+    def links(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support link checks.' % self.language)
+
+    def run(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support run checks.' % self.language)
+
+    def sizeof(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support sizeof checks.' % self.language)
+
+    def alignment(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support alignment checks.' % self.language)
+
+    def has_function(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support function checks.' % self.language)
+
+class CCompiler(Compiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version)
         self.language = 'c'
         self.default_suffix = 'c'
         self.id = 'unknown'
@@ -142,9 +190,6 @@ class CCompiler():
     def get_always_args(self):
         return []
 
-    def get_linker_always_args(self):
-        return []
-
     def get_warn_args(self, level):
         return self.warn_args[level]
 
@@ -202,6 +247,8 @@ class CCompiler():
         return []
 
     def get_include_args(self, path):
+        if path == '':
+            path = '.'
         return ['-I' + path]
 
     def get_std_shared_lib_link_args(self):
@@ -238,8 +285,26 @@ class CCompiler():
         ofile = open(source_name, 'w')
         ofile.write('int main(int argc, char **argv) { int class=0; return class; }\n')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
-        pc.wait()
+        if self.is_cross and self.exe_wrapper is None:
+            # Linking cross built apps is painful. You can't really
+            # tell if you should use -nostdlib or not and for example
+            # on OSX the compiler binary is the same but you need
+            # a ton of compiler flags to differentiate between
+            # arm and x86_64. So just compile.
+            extra_flags = ['-c']
+        else:
+            extra_flags = []
+        cmdlist = self.exelist + extra_flags + [source_name, '-o', binary_name]
+        pc = subprocess.Popen(cmdlist, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        (stdo, stde) = pc.communicate()
+        stdo = stdo.decode()
+        stde = stde.decode()
+        mlog.debug('Sanity check compiler command line:', ' '.join(cmdlist))
+        mlog.debug('Sanity check compile stdout:')
+        mlog.debug(stdo)
+        mlog.debug('-----\nSanity check compile stderr:')
+        mlog.debug(stde)
+        mlog.debug('-----')
         if pc.returncode != 0:
             raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
         if self.is_cross:
@@ -255,24 +320,17 @@ class CCompiler():
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by C compiler %s are not runnable.' % self.name_string())
 
-    def has_header(self, hname):
+    def has_header(self, hname, extra_args=[]):
         templ = '''#include<%s>
 int someSymbolHereJustForFun;
 '''
-        return self.compiles(templ % hname)
+        return self.compiles(templ % hname, extra_args)
 
-    def compiles(self, code, extra_args = []):
-        suflen = len(self.default_suffix)
-        (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
-        os.close(fd)
-        ofile = open(srcname, 'w')
-        ofile.write(code)
-        ofile.close()
+    def compile(self, code, srcname, extra_args=[]):
         commands = self.get_exelist()
-        commands += extra_args
-        commands += self.get_compile_only_args()
         commands.append(srcname)
-        mlog.debug('Running compile test.')
+        commands += extra_args
+        mlog.debug('Running compile:')
         mlog.debug('Command line: ', ' '.join(commands))
         mlog.debug('Code:\n', code)
         p = subprocess.Popen(commands, cwd=os.path.split(srcname)[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
@@ -282,6 +340,17 @@ int someSymbolHereJustForFun;
         mlog.debug('Compiler stdout:\n', stdo)
         mlog.debug('Compiler stderr:\n', stde)
         os.remove(srcname)
+        return p
+
+    def compiles(self, code, extra_args = []):
+        suflen = len(self.default_suffix)
+        (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
+        os.close(fd)
+        ofile = open(srcname, 'w')
+        ofile.write(code)
+        ofile.close()
+        extra_args = extra_args + self.get_compile_only_args()
+        p = self.compile(code, srcname, extra_args)
         try:
             trial = srcname[:-suflen] + 'o'
             os.remove(trial)
@@ -293,7 +362,24 @@ int someSymbolHereJustForFun;
             pass
         return p.returncode == 0
 
-    def run(self, code):
+    def links(self, code, extra_args = []):
+        suflen = len(self.default_suffix)
+        (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
+        os.close(fd)
+        (fd, dstname) = tempfile.mkstemp()
+        os.close(fd)
+        ofile = open(srcname, 'w')
+        ofile.write(code)
+        ofile.close()
+        extra_args = extra_args + self.get_output_args(dstname)
+        p = self.compile(code, srcname, extra_args)
+        try:
+            os.remove(dstname)
+        except FileNotFoundError:
+            pass
+        return p.returncode == 0
+
+    def run(self, code, extra_args=[]):
         mlog.debug('Running code:\n\n', code)
         if self.is_cross and self.exe_wrapper is None:
             raise CrossNoRunException('Can not run test applications in this cross environment.')
@@ -304,6 +390,7 @@ int someSymbolHereJustForFun;
         ofile.close()
         exename = srcname + '.exe' # Is guaranteed to be executable on every platform.
         commands = self.get_exelist()
+        commands += extra_args
         commands.append(srcname)
         commands += self.get_output_args(exename)
         p = subprocess.Popen(commands, cwd=os.path.split(srcname)[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
@@ -328,13 +415,12 @@ int someSymbolHereJustForFun;
         os.remove(exename)
         return RunResult(True, pe.returncode, so, se)
 
-    def cross_sizeof(self, element, prefix, env):
+    def cross_sizeof(self, element, prefix, env, extra_args=[]):
         templ = '''%s
 int temparray[%d-sizeof(%s)];
 '''
-        extra_args = []
         try:
-            extra_args = env.cross_info.config['properties'][self.language + '_args']
+            extra_args += env.cross_info.config['properties'][self.language + '_args']
         except KeyError:
             pass
         for i in range(1, 1024):
@@ -343,9 +429,9 @@ int temparray[%d-sizeof(%s)];
                 return i
         raise EnvironmentException('Cross checking sizeof overflowed.')
 
-    def sizeof(self, element, prefix, env):
+    def sizeof(self, element, prefix, env, extra_args=[]):
         if self.is_cross:
-            return self.cross_sizeof(element, prefix, env)
+            return self.cross_sizeof(element, prefix, env, extra_args)
         templ = '''#include<stdio.h>
 %s
 
@@ -354,14 +440,14 @@ int main(int argc, char **argv) {
     return 0;
 };
 '''
-        res = self.run(templ % (prefix, element))
+        res = self.run(templ % (prefix, element), extra_args)
         if not res.compiled:
             raise EnvironmentException('Could not compile sizeof test.')
         if res.returncode != 0:
             raise EnvironmentException('Could not run sizeof test binary.')
         return int(res.stdout)
 
-    def cross_alignment(self, typename, env):
+    def cross_alignment(self, typename, env, extra_args=[]):
         templ = '''#include<stddef.h>
 struct tmp {
   char c;
@@ -371,7 +457,7 @@ struct tmp {
 int testarray[%d-offsetof(struct tmp, target)];
 '''
         try:
-            extra_args = env.cross_info.config['properties'][self.language + '_args']
+            extra_args += env.cross_info.config['properties'][self.language + '_args']
         except KeyError:
             pass
         for i in range(1, 1024):
@@ -380,9 +466,9 @@ int testarray[%d-offsetof(struct tmp, target)];
                 return i
         raise EnvironmentException('Cross checking offsetof overflowed.')
 
-    def alignment(self, typename, env):
+    def alignment(self, typename, env, extra_args=[]):
         if self.is_cross:
-            return self.cross_alignment(typename, env)
+            return self.cross_alignment(typename, env, extra_args)
         templ = '''#include<stdio.h>
 #include<stddef.h>
 
@@ -396,7 +482,7 @@ int main(int argc, char **argv) {
   return 0;
 }
 '''
-        res = self.run(templ % typename)
+        res = self.run(templ % typename, extra_args)
         if not res.compiled:
             raise EnvironmentException('Could not compile alignment test.')
         if res.returncode != 0:
@@ -406,7 +492,7 @@ int main(int argc, char **argv) {
             raise EnvironmentException('Could not determine alignment of %s. Sorry. You might want to file a bug.' % typename)
         return align
 
-    def has_function(self, funcname, prefix, env):
+    def has_function(self, funcname, prefix, env, extra_args=[]):
         # This fails (returns true) if funcname is a ptr or a variable.
         # The correct check is a lot more difficult.
         # Fix this to do that eventually.
@@ -424,24 +510,24 @@ int main(int argc, char **argv) {
                 if isinstance(val, bool):
                     return val
                 raise EnvironmentException('Cross variable {0} is not a boolean.'.format(varname))
-        return self.compiles(templ % (prefix, funcname))
+        return self.compiles(templ % (prefix, funcname), extra_args)
 
-    def has_member(self, typename, membername, prefix):
+    def has_member(self, typename, membername, prefix, extra_args=[]):
         templ = '''%s
 void bar() {
     %s foo;
     foo.%s;
 };
 '''
-        return self.compiles(templ % (prefix, typename, membername))
+        return self.compiles(templ % (prefix, typename, membername), extra_args)
 
-    def has_type(self, typename, prefix):
+    def has_type(self, typename, prefix, extra_args):
         templ = '''%s
 void bar() {
     sizeof(%s);
 };
 '''
-        return self.compiles(templ % (prefix, typename))
+        return self.compiles(templ % (prefix, typename), extra_args)
 
     def thread_flags(self):
         return ['-pthread']
@@ -467,7 +553,23 @@ class CPPCompiler(CCompiler):
         ofile = open(source_name, 'w')
         ofile.write('class breakCCompiler;int main(int argc, char **argv) { return 0; }\n')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
+        if self.is_cross and self.exe_wrapper is None:
+            # Skipping link because of the same reason as for C.
+            # The comment in CCompiler explains why this is done.
+            extra_flags = ['-c']
+        else:
+            extra_flags = []
+        cmdlist = self.exelist + extra_flags + [source_name, '-o', binary_name]
+        pc = subprocess.Popen(cmdlist, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        (stdo, stde) = pc.communicate()
+        stdo = stdo.decode()
+        stde = stde.decode()
+        mlog.debug('Sanity check compiler command line:', ' '.join(cmdlist))
+        mlog.debug('Sanity check compile stdout:')
+        mlog.debug(stdo)
+        mlog.debug('-----\nSanity check compile stderr:')
+        mlog.debug(stde)
+        mlog.debug('-----')
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
@@ -537,29 +639,17 @@ class ObjCPPCompiler(CPPCompiler):
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by ObjC++ compiler %s are not runnable.' % self.name_string())
 
-class MonoCompiler():
+class MonoCompiler(Compiler):
     def __init__(self, exelist, version):
-        if type(exelist) == type(''):
-            self.exelist = [exelist]
-        elif type(exelist) == type([]):
-            self.exelist = exelist
-        else:
-            raise TypeError('Unknown argument to Mono compiler')
-        self.version = version
+        super().__init__(exelist, version)
         self.language = 'cs'
         self.default_suffix = 'cs'
         self.id = 'mono'
         self.monorunner = 'mono'
 
-    def get_always_args(self):
-        return []
-
     def get_output_args(self, fname):
         return ['-out:' + fname]
 
-    def get_linker_always_args(self):
-        return []
-
     def get_link_args(self, fname):
         return ['-r:' + fname]
 
@@ -656,47 +746,17 @@ class MonoCompiler():
     def needs_static_linker(self):
         return False
 
-    def has_header(self, hname):
-        raise EnvironmentException('Mono does not support header checks.')
-
-    def compiles(self, code):
-        raise EnvironmentException('Mono does not support compile checks.')
-
-    def run(self, code):
-        raise EnvironmentException('Mono does not support run checks.')
-
-    def sizeof(self, element, prefix, env):
-        raise EnvironmentException('Mono does not support sizeof checks.')
-
-    def alignment(self, typename, env):
-        raise EnvironmentException('Mono does not support alignment checks.')
-
-    def has_function(self, funcname, prefix, env):
-        raise EnvironmentException('Mono does not support function checks.')
-
     def get_buildtype_args(self, buildtype):
         return mono_buildtype_args[buildtype]
 
-class JavaCompiler():
+class JavaCompiler(Compiler):
     def __init__(self, exelist, version):
-        if type(exelist) == type(''):
-            self.exelist = [exelist]
-        elif type(exelist) == type([]):
-            self.exelist = exelist
-        else:
-            raise TypeError('Unknown argument to JavaCompiler')
-        self.version = version
+        super().__init__(exelist, version)
         self.language = 'java'
         self.default_suffix = 'java'
         self.id = 'unknown'
         self.javarunner = 'java'
 
-    def get_always_args(self):
-        return []
-
-    def get_linker_always_args(self):
-        return []
-
     def get_soname_args(self, shlib_name, path, soversion):
         return []
 
@@ -799,32 +859,9 @@ class JavaCompiler():
     def needs_static_linker(self):
         return False
 
-    def has_header(self, hname):
-        raise EnvironmentException('Java does not support header checks.')
-
-    def compiles(self, code):
-        raise EnvironmentException('Java does not support compile checks.')
-
-    def run(self, code):
-        raise EnvironmentException('Java does not support run checks.')
-
-    def sizeof(self, element, prefix, env):
-        raise EnvironmentException('Java does not support sizeof checks.')
-
-    def alignment(self, typename, env):
-        raise EnvironmentException('Java does not support alignment checks.')
-
-    def has_function(self, funcname, prefix, env):
-        raise EnvironmentException('Java does not support function checks.')
-
-class ValaCompiler():
+class ValaCompiler(Compiler):
     def __init__(self, exelist, version):
-        if isinstance(exelist, str):
-            self.exelist = [exelist]
-        elif type(exelist) == type([]):
-            self.exelist = exelist
-        else:
-            raise TypeError('Unknown argument to Vala compiler')
+        super().__init__(exelist, version)
         self.version = version
         self.id = 'unknown'
         self.language = 'vala'
@@ -861,15 +898,9 @@ class ValaCompiler():
         suffix = filename.split('.')[-1]
         return suffix in ('vala', 'vapi')
 
-class RustCompiler():
+class RustCompiler(Compiler):
     def __init__(self, exelist, version):
-        if isinstance(exelist, str):
-            self.exelist = [exelist]
-        elif type(exelist) == type([]):
-            self.exelist = exelist
-        else:
-            raise TypeError('Unknown argument to Rust compiler')
-        self.version = version
+        super().__init__(exelist, version)
         self.id = 'unknown'
         self.language = 'rust'
 
@@ -918,6 +949,7 @@ class VisualStudioCCompiler(CCompiler):
     vs2010_always_args = ['/nologo', '/showIncludes']
     vs2013_always_args = ['/nologo', '/showIncludes', '/FS']
 
+
     def __init__(self, exelist, version, is_cross, exe_wrap):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.id = 'msvc'
@@ -1010,6 +1042,15 @@ class VisualStudioCCompiler(CCompiler):
     def thread_link_flags(self):
         return []
 
+    def get_options(self):
+        return {'c_winlibs' : coredata.UserStringArrayOption('c_winlibs',
+                                                             'Windows libs to link against.',
+                                                             msvc_winlibs)
+                }
+
+    def get_option_link_args(self, options):
+        return options['c_winlibs'].value
+
 class VisualStudioCPPCompiler(VisualStudioCCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrap):
         VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
@@ -1040,6 +1081,26 @@ class VisualStudioCPPCompiler(VisualStudioCCompiler):
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by C++ compiler %s are not runnable.' % self.name_string())
 
+    def get_options(self):
+        return {'cpp_eh' : coredata.UserComboOption('cpp_eh',
+                                                    'C++ exception handling type.',
+                                                    ['none', 'a', 's', 'sc'],
+                                                    'sc'),
+                'cpp_winlibs' : coredata.UserStringArrayOption('cpp_winlibs',
+                                                               'Windows libs to link against.',
+                                                               msvc_winlibs)
+                }
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['cpp_eh']
+        if std.value != 'none':
+            args.append('/EH' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        return options['cpp_winlibs'].value
+
 GCC_STANDARD = 0
 GCC_OSX = 1
 GCC_MINGW = 2
@@ -1057,13 +1118,14 @@ def get_gcc_soname_args(gcc_type, shlib_name, path, soversion):
     else:
         raise RuntimeError('Not implemented yet.')
 
+
 class GnuCCompiler(CCompiler):
     def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'gcc'
         self.gcc_type = gcc_type
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
 
     def get_pic_args(self):
@@ -1092,6 +1154,29 @@ class GnuCCompiler(CCompiler):
     def can_compile(self, filename):
         return super().can_compile(filename) or filename.split('.')[-1].lower() == 's' # Gcc can do asm, too.
 
+    def get_options(self):
+        opts = {'c_std' : coredata.UserComboOption('c_std', 'C language standard to use',
+                                                   ['none', 'c89', 'c99', 'c11', 'gnu89', 'gnu99', 'gnu11'],
+                                                   'c11')}
+        if self.gcc_type == GCC_MINGW:
+            opts.update({
+                'c_winlibs': coredata.UserStringArrayOption('c_winlibs', 'Standard Win libraries to link against',
+                                                            gnu_winlibs),
+                })
+        return opts
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['c_std']
+        if std.value != 'none':
+            args.append('-std=' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        if self.gcc_type == GCC_MINGW:
+            return options['c_winlibs'].value
+        return []
+
 class GnuObjCCompiler(ObjCCompiler):
     std_opt_args = ['-O2']
 
@@ -1102,7 +1187,7 @@ class GnuObjCCompiler(ObjCCompiler):
         # if this breaks your use case.
         self.gcc_type = GCC_STANDARD
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
 
     def get_buildtype_args(self, buildtype):
@@ -1127,7 +1212,7 @@ class GnuObjCPPCompiler(ObjCPPCompiler):
         # if this breaks your use case.
         self.gcc_type = GCC_STANDARD
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch', '-Wnon-virtual-dtor'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
 
     def get_buildtype_args(self, buildtype):
@@ -1153,13 +1238,11 @@ class ClangObjCPPCompiler(GnuObjCPPCompiler):
         self.id = 'clang'
 
 class ClangCCompiler(CCompiler):
-    std_warn_args = ['-Wall', '-Wpedantic', '-Winvalid-pch']
-
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Weverything']}
 
     def get_buildtype_args(self, buildtype):
@@ -1180,6 +1263,20 @@ class ClangCCompiler(CCompiler):
         # so it might change semantics at any time.
         return ['-include-pch', os.path.join (pch_dir, self.get_pch_name (header))]
 
+    def get_options(self):
+        return {'c_std' : coredata.UserComboOption('c_std', 'C language standard to use',
+                                                   ['none', 'c89', 'c99', 'c11'],
+                                                   'c11')}
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['c_std']
+        if std.value != 'none':
+            args.append('-std=' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        return []
 
 class GnuCPPCompiler(CPPCompiler):
     # may need to separate the latter to extra_debug_args or something
@@ -1189,8 +1286,8 @@ class GnuCPPCompiler(CPPCompiler):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.id = 'gcc'
         self.gcc_type = gcc_type
-        self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch', '-Wnon-virtual-dtor'],
+        self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3': ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
 
     def get_always_args(self):
@@ -1208,12 +1305,35 @@ class GnuCPPCompiler(CPPCompiler):
     def get_soname_args(self, shlib_name, path, soversion):
         return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
+    def get_options(self):
+        opts = {'cpp_std' : coredata.UserComboOption('cpp_std', 'C language standard to use',
+                                                     ['none', 'c++03', 'c++11', 'c++1y'],
+                                                     'c++11')}
+        if self.gcc_type == GCC_MINGW:
+            opts.update({
+                'cpp_winlibs': coredata.UserStringArrayOption('c_winlibs', 'Standard Win libraries to link against',
+                                                              gnu_winlibs),
+                })
+        return opts
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['cpp_std']
+        if std.value != 'none':
+            args.append('-std=' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        if self.gcc_type == GCC_MINGW:
+            return options['cpp_winlibs'].value
+        return []
+
 class ClangCPPCompiler(CPPCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
-                          '2': ['-Wall', '-Wpedantic', '-Winvalid-pch', '-Wnon-virtual-dtor'],
+                          '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3': ['-Weverything']}
 
     def get_buildtype_args(self, buildtype):
@@ -1231,11 +1351,24 @@ class ClangCPPCompiler(CPPCompiler):
         # so it might change semantics at any time.
         return ['-include-pch', os.path.join (pch_dir, self.get_pch_name (header))]
 
-class FortranCompiler():
-    def __init__(self, exelist, version,is_cross, exe_wrapper=None):
-        super().__init__()
-        self.exelist = exelist
-        self.version = version
+    def get_options(self):
+        return {'cpp_std' : coredata.UserComboOption('cpp_std', 'C++ language standard to use',
+                                                   ['none', 'c++03', 'c++11', 'c++1y'],
+                                                   'c++11')}
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['cpp_std']
+        if std.value != 'none':
+            args.append('-std=' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        return []
+
+class FortranCompiler(Compiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version)
         self.is_cross = is_cross
         self.exe_wrapper = exe_wrapper
         self.language = 'fortran'
@@ -1291,12 +1424,6 @@ end program prog
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by Fortran compiler %s are not runnable.' % self.name_string())
 
-    def get_always_args(self):
-        return ['-pipe']
-
-    def get_linker_always_args(self):
-        return []
-
     def get_std_warn_args(self, level):
         return FortranCompiler.std_warn_args
 
@@ -1366,6 +1493,9 @@ class GnuFortranCompiler(FortranCompiler):
         self.gcc_type = gcc_type
         self.id = 'gcc'
 
+    def get_always_args(self):
+        return ['-pipe']
+
 class G95FortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper=None)
@@ -1374,6 +1504,9 @@ class G95FortranCompiler(FortranCompiler):
     def get_module_outdir_args(self, path):
         return ['-fmod='+path]
 
+    def get_always_args(self):
+        return ['-pipe']
+
 class SunFortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper=None)
@@ -1401,9 +1534,6 @@ class IntelFortranCompiler(FortranCompiler):
     def get_module_outdir_args(self, path):
         return ['-module', path]
 
-    def get_always_args(self):
-        return []
-
     def can_compile(self, src):
         suffix = os.path.splitext(src)[1].lower()
         if suffix == '.f' or suffix == '.f90':
@@ -1423,9 +1553,6 @@ class PathScaleFortranCompiler(FortranCompiler):
     def get_module_outdir_args(self, path):
         return ['-module', path]
 
-    def get_always_args(self):
-        return []
-
     def can_compile(self, src):
         suffix = os.path.splitext(src)[1].lower()
         if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
@@ -1445,9 +1572,6 @@ class PGIFortranCompiler(FortranCompiler):
     def get_module_outdir_args(self, path):
         return ['-module', path]
 
-    def get_always_args(self):
-        return []
-
     def can_compile(self, src):
         suffix = os.path.splitext(src)[1].lower()
         if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
@@ -1468,9 +1592,6 @@ class Open64FortranCompiler(FortranCompiler):
     def get_module_outdir_args(self, path):
         return ['-module', path]
 
-    def get_always_args(self):
-        return []
-
     def can_compile(self, src):
         suffix = os.path.splitext(src)[1].lower()
         if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
@@ -1535,6 +1656,9 @@ class VisualStudioLinker():
     def thread_link_flags(self):
         return []
 
+    def get_option_link_args(self, options):
+        return []
+
 class ArLinker():
     std_args = ['csr']
 
@@ -1568,3 +1692,6 @@ class ArLinker():
 
     def thread_link_flags(self):
         return []
+
+    def get_option_link_args(self, options):
+        return []
