@@ -59,11 +59,12 @@ class Dependency():
         return False
 
 class InternalDependency():
-    def __init__(self, incdirs, libraries, sources):
+    def __init__(self, incdirs, libraries, sources, ext_deps):
         super().__init__()
         self.include_directories = incdirs
         self.libraries = libraries
         self.sources = sources
+        self.ext_deps = ext_deps
 
 class PkgConfigDependency(Dependency):
     pkgconfig_found = None
@@ -95,6 +96,7 @@ class PkgConfigDependency(Dependency):
             pkgbin = 'pkg-config'
             self.type_string = 'Native'
 
+        mlog.debug('Determining dependency %s with pkg-config executable %s.' % (name, pkgbin))
         self.pkgbin = pkgbin
         p = subprocess.Popen([pkgbin, '--modversion', name],
                              stdout=subprocess.PIPE,
