@@ -12,9 +12,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import re
-import coredata
-from glob import glob
+import os, re, subprocess
+import coredata, mesonlib
 from compilers import *
 import configparser
 
@@ -68,13 +67,15 @@ class Environment():
         try:
             cdf = os.path.join(self.get_build_dir(), Environment.coredata_file)
             self.coredata = coredata.load(cdf)
+            self.first_invocation = False
         except FileNotFoundError:
             self.coredata = coredata.CoreData(options)
+            self.first_invocation = True
         if self.coredata.cross_file:
             self.cross_info = CrossBuildInfo(self.coredata.cross_file)
         else:
             self.cross_info = None
-        self.cmd_line_options = options.projectoptions
+        self.cmd_line_options = options
 
         # List of potential compilers.
         if mesonlib.is_windows():
@@ -542,22 +543,22 @@ class Environment():
         return self.object_suffix
 
     def get_prefix(self):
-        return self.coredata.prefix
+        return self.coredata.get_builtin_option('prefix')
 
     def get_libdir(self):
-        return self.coredata.libdir
+        return self.coredata.get_builtin_option('libdir')
 
     def get_bindir(self):
-        return self.coredata.bindir
+        return self.coredata.get_builtin_option('bindir')
 
     def get_includedir(self):
-        return self.coredata.includedir
+        return self.coredata.get_builtin_option('includedir')
 
     def get_mandir(self):
-        return self.coredata.mandir
+        return self.coredata.get_builtin_option('mandir')
 
     def get_datadir(self):
-        return self.coredata.datadir
+        return self.coredata.get_builtin_option('datadir')
 
     def find_library(self, libname, dirs):
         if dirs is None:
