@@ -39,8 +39,10 @@ class InvalidArguments(InterpreterException):
 
 def check_stringlist(a, msg='Arguments must be strings.'):
     if not isinstance(a, list):
+        mlog.debug('Not a list:', str(a))
         raise InvalidArguments('Argument not a list.')
     if not all(isinstance(s, str) for s in a):
+        mlog.debug('Element not a string:', str(a))
         raise InvalidArguments(msg)
 
 def noPosargs(f):
@@ -174,6 +176,7 @@ class ConfigurationDataHolder(InterpreterObject):
         self.held_object = build.ConfigurationData()
         self.methods.update({'set': self.set_method,
                              'set10': self.set10_method,
+                             'has' : self.has_method,
                             })
 
     def is_used(self):
@@ -204,6 +207,9 @@ class ConfigurationDataHolder(InterpreterObject):
         else:
             self.held_object.values[name] = 0
 
+    def has_method(self, args, kwargs):
+        return args[0] in self.held_object.values
+
     def get(self, name):
         return self.held_object.values[name]
 
@@ -306,6 +312,7 @@ class BuildMachine(InterpreterObject):
     def __init__(self):
         InterpreterObject.__init__(self)
         self.methods.update({'system' : self.system_method,
+                             'cpu_family' : self.cpu_family_method,
                              'cpu' : self.cpu_method,
                              'endian' : self.endian_method,
                             })
@@ -314,18 +321,18 @@ class BuildMachine(InterpreterObject):
     # It returns different values for the same cpu.
     # For x86 it might return 'x86', 'i686' or somesuch.
     # Do some canonicalization.
-    def cpu_method(self, args, kwargs):
+    def cpu_family_method(self, args, kwargs):
         trial = platform.machine().lower()
         if trial.startswith('i') and trial.endswith('86'):
             return 'x86'
-        # This might be wrong. Maybe we should return the more
-        # specific string such as 'armv7l'. Need to get user
-        # feedback first.
         if trial.startswith('arm'):
             return 'arm'
         # Add fixes here as bugs are reported.
         return trial
 
+    def cpu_method(self, args, kwargs):
+        return platform.machine().lower()
+
     def system_method(self, args, kwargs):
         return platform.system().lower()
 
@@ -337,9 +344,15 @@ class BuildMachine(InterpreterObject):
 class CrossMachineInfo(InterpreterObject):
     def __init__(self, cross_info):
         InterpreterObject.__init__(self)
+        minimum_cross_info = {'cpu', 'cpu_family', 'endian', 'system'}
+        if set(cross_info) < minimum_cross_info:
+            raise InterpreterException(
+                'Machine info is currently {}\n'.format(cross_info) +
+                'but is missing {}.'.format(minimum_cross_info - set(cross_info)))
         self.info = cross_info
         self.methods.update({'system' : self.system_method,
                              'cpu' : self.cpu_method,
+                             'cpu_family' : self.cpu_family_method,
                              'endian' : self.endian_method,
                             })
 
@@ -349,13 +362,16 @@ class CrossMachineInfo(InterpreterObject):
     def cpu_method(self, args, kwargs):
         return self.info['cpu']
 
+    def cpu_family_method(self, args, kwargs):
+        return self.info['cpu_family']
+
     def endian_method(self, args, kwargs):
         return self.info['endian']
 
 class IncludeDirsHolder(InterpreterObject):
-    def __init__(self, curdir, dirs):
+    def __init__(self, idobj):
         super().__init__()
-        self.held_object = build.IncludeDirs(curdir, dirs)
+        self.held_object = idobj
 
 class Headers(InterpreterObject):
 
@@ -449,15 +465,27 @@ class GeneratedObjectsHolder(InterpreterObject):
         self.held_object = held_object
 
 class BuildTargetHolder(InterpreterObject):
-    def __init__(self, target):
+    def __init__(self, target, interp):
         super().__init__()
         self.held_object = target
+        self.interpreter = interp
         self.methods.update({'extract_objects' : self.extract_objects_method,
-                             'extract_all_objects' : self.extract_all_objects_method})
+                             'extract_all_objects' : self.extract_all_objects_method,
+                             'get_id': self.get_id_method,
+                             'outdir' : self.outdir_method,
+                             'private_dir_include' : self.private_dir_include_method,
+                             })
 
     def is_cross(self):
         return self.held_object.is_cross()
 
+    def private_dir_include_method(self, args, kwargs):
+        return IncludeDirsHolder(build.IncludeDirs('', [],
+                    [self.interpreter.backend.get_target_private_dir(self.held_object)]))
+
+    def outdir_method(self, args, kwargs):
+        return self.interpreter.backend.get_target_dir(self.held_object)
+
     def extract_objects_method(self, args, kwargs):
         gobjs = self.held_object.extract_objects(args)
         return GeneratedObjectsHolder(gobjs)
@@ -466,21 +494,24 @@ class BuildTargetHolder(InterpreterObject):
         gobjs = self.held_object.extract_all_objects()
         return GeneratedObjectsHolder(gobjs)
 
+    def get_id_method(self, args, kwargs):
+        return self.held_object.get_id()
+
 class ExecutableHolder(BuildTargetHolder):
-    def __init__(self, target):
-        super().__init__(target)
+    def __init__(self, target, interp):
+        super().__init__(target, interp)
 
 class StaticLibraryHolder(BuildTargetHolder):
-    def __init__(self, target):
-        super().__init__(target)
+    def __init__(self, target, interp):
+        super().__init__(target, interp)
 
 class SharedLibraryHolder(BuildTargetHolder):
-    def __init__(self, target):
-        super().__init__(target)
+    def __init__(self, target, interp):
+        super().__init__(target, interp)
 
 class JarHolder(BuildTargetHolder):
-    def __init__(self, target):
-        super().__init__(target)
+    def __init__(self, target, interp):
+        super().__init__(target, interp)
 
 class CustomTargetHolder(InterpreterObject):
     def __init__(self, object_to_hold):
@@ -537,6 +568,7 @@ class CompilerHolder(InterpreterObject):
         self.compiler = compiler
         self.environment = env
         self.methods.update({'compiles': self.compiles_method,
+                             'links': self.links_method,
                              'get_id': self.get_id_method,
                              'sizeof': self.sizeof_method,
                              'has_header': self.has_header_method,
@@ -555,12 +587,25 @@ class CompilerHolder(InterpreterObject):
     def cmd_array_method(self, args, kwargs):
         return self.compiler.exelist
 
+    def determine_args(self, kwargs):
+        nobuiltins = kwargs.get('no_builtin_args', False)
+        if not isinstance(nobuiltins, bool):
+            raise InterpreterException('Type of no_builtin_args not a boolean.')
+        args = []
+        if not nobuiltins:
+            opts = self.environment.coredata.compiler_options
+            args += self.compiler.get_option_compile_args(opts)
+            args += self.compiler.get_option_link_args(opts)
+        args += mesonlib.stringlistify(kwargs.get('args', []))
+        return args
+
     def alignment_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Alignment method takes exactly one positional argument.')
         check_stringlist(args)
         typename = args[0]
-        result = self.compiler.alignment(typename, self.environment)
+        extra_args = mesonlib.stringlistify(kwargs.get('args', []))
+        result = self.compiler.alignment(typename, self.environment, extra_args)
         mlog.log('Checking for alignment of "', mlog.bold(typename), '": ', result, sep='')
         return result
 
@@ -572,7 +617,8 @@ class CompilerHolder(InterpreterObject):
         testname = kwargs.get('name', '')
         if not isinstance(testname, str):
             raise InterpreterException('Testname argument must be a string.')
-        result = self.compiler.run(code)
+        extra_args = self.determine_args(kwargs)
+        result = self.compiler.run(code, extra_args)
         if len(testname) > 0:
             if not result.compiled:
                 h = mlog.red('DID NOT COMPILE')
@@ -595,7 +641,8 @@ class CompilerHolder(InterpreterObject):
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of has_function must be a string.')
-        had = self.compiler.has_member(typename, membername, prefix)
+        extra_args = self.determine_args(kwargs)
+        had = self.compiler.has_member(typename, membername, prefix, extra_args)
         if had:
             hadtxt = mlog.green('YES')
         else:
@@ -612,7 +659,8 @@ class CompilerHolder(InterpreterObject):
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of has_function must be a string.')
-        had = self.compiler.has_function(funcname, prefix, self.environment)
+        extra_args = self.determine_args(kwargs)
+        had = self.compiler.has_function(funcname, prefix, self.environment, extra_args)
         if had:
             hadtxt = mlog.green('YES')
         else:
@@ -628,7 +676,8 @@ class CompilerHolder(InterpreterObject):
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of has_type must be a string.')
-        had = self.compiler.has_type(typename, prefix)
+        extra_args = self.determine_args(kwargs)
+        had = self.compiler.has_type(typename, prefix, extra_args)
         if had:
             hadtxt = mlog.green('YES')
         else:
@@ -644,7 +693,8 @@ class CompilerHolder(InterpreterObject):
         prefix = kwargs.get('prefix', '')
         if not isinstance(prefix, str):
             raise InterpreterException('Prefix argument of sizeof must be a string.')
-        esize = self.compiler.sizeof(element, prefix, self.environment)
+        extra_args = self.determine_args(kwargs)
+        esize = self.compiler.sizeof(element, prefix, self.environment, extra_args)
         mlog.log('Checking for size of "%s": %d' % (element, esize))
         return esize
 
@@ -652,11 +702,12 @@ class CompilerHolder(InterpreterObject):
         if len(args) != 1:
             raise InterpreterException('compiles method takes exactly one argument.')
         check_stringlist(args)
-        string = args[0]
+        code = args[0]
         testname = kwargs.get('name', '')
         if not isinstance(testname, str):
             raise InterpreterException('Testname argument must be a string.')
-        result = self.compiler.compiles(string)
+        extra_args = self.determine_args(kwargs)
+        result = self.compiler.compiles(code, extra_args)
         if len(testname) > 0:
             if result:
                 h = mlog.green('YES')
@@ -665,12 +716,31 @@ class CompilerHolder(InterpreterObject):
             mlog.log('Checking if "', mlog.bold(testname), '" compiles : ', h, sep='')
         return result
 
+    def links_method(self, args, kwargs):
+        if len(args) != 1:
+            raise InterpreterException('links method takes exactly one argument.')
+        check_stringlist(args)
+        code = args[0]
+        testname = kwargs.get('name', '')
+        if not isinstance(testname, str):
+            raise InterpreterException('Testname argument must be a string.')
+        extra_args = self.determine_args(kwargs)
+        result = self.compiler.links(code, extra_args)
+        if len(testname) > 0:
+            if result:
+                h = mlog.green('YES')
+            else:
+                h = mlog.red('NO')
+            mlog.log('Checking if "', mlog.bold(testname), '" links : ', h, sep='')
+        return result
+
     def has_header_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('has_header method takes exactly one argument.')
         check_stringlist(args)
         string = args[0]
-        haz = self.compiler.has_header(string)
+        extra_args = self.determine_args(kwargs)
+        haz = self.compiler.has_header(string, extra_args)
         if haz:
             h = mlog.green('YES')
         else:
@@ -789,7 +859,7 @@ class MesonMain(InterpreterObject):
         raise InterpreterException('Tried to access compiler for unspecified language "%s".' % cname)
 
     def is_unity_method(self, args, kwargs):
-        return self.build.environment.coredata.unity
+        return self.build.environment.coredata.get_builtin_option('unity')
 
     def is_subproject_method(self, args, kwargs):
         return self.interpreter.is_subproject()
@@ -806,8 +876,9 @@ class MesonMain(InterpreterObject):
 
 class Interpreter():
 
-    def __init__(self, build, subproject='', subdir='', subproject_dir='subprojects'):
+    def __init__(self, build, backend, subproject='', subdir='', subproject_dir='subprojects'):
         self.build = build
+        self.backend = backend
         self.subproject = subproject
         self.subdir = subdir
         self.source_root = build.environment.get_source_dir()
@@ -815,7 +886,7 @@ class Interpreter():
         option_file = os.path.join(self.source_root, self.subdir, 'meson_options.txt')
         if os.path.exists(option_file):
             oi = optinterpreter.OptionInterpreter(self.subproject, \
-                                                  self.build.environment.cmd_line_options)
+                                                  self.build.environment.cmd_line_options.projectoptions)
             oi.process(option_file)
             self.build.environment.merge_options(oi.options)
         mesonfile = os.path.join(self.source_root, self.subdir, environment.build_filename)
@@ -866,6 +937,7 @@ class Interpreter():
                       'dependency' : self.func_dependency,
                       'static_library' : self.func_static_lib,
                       'shared_library' : self.func_shared_lib,
+                      'library' : self.func_library,
                       'jar' : self.func_jar,
                       'build_target': self.func_build_target,
                       'custom_target' : self.func_custom_target,
@@ -892,9 +964,12 @@ class Interpreter():
                       'pkgconfig_gen' : self.func_pkgconfig_gen,
                       'vcs_tag' : self.func_vcs_tag,
                       'set_variable' : self.func_set_variable,
+                      'is_variable' : self.func_is_variable,
+                      'get_variable' : self.func_get_variable,
                       'import' : self.func_import,
                       'files' : self.func_files,
                       'declare_dependency': self.func_declare_dependency,
+                      'assert': self.func_assert,
                      }
 
     def module_method_callback(self, invalues):
@@ -990,6 +1065,29 @@ class Interpreter():
         value = self.to_native(args[1])
         self.set_variable(varname, value)
 
+    @noKwargs
+    def func_get_variable(self, node, args, kwargs):
+        if len(args)<1 or len(args)>2:
+            raise InvalidCode('Get_variable takes one or two arguments.')
+        varname = args[0]
+        if not isinstance(varname, str):
+            raise InterpreterException('First argument must be a string.')
+        try:
+            return self.variables[varname]
+        except KeyError:
+            pass
+        if len(args) == 2:
+            return args[1]
+        raise InterpreterException('Tried to get unknown variable "%s".' % varname)
+
+    @stringArgs
+    @noKwargs
+    def func_is_variable(self, node, args, kwargs):
+        if len(args) != 1:
+            raise InvalidCode('Is_variable takes two arguments.')
+        varname = args[0]
+        return varname in self.variables
+
     @stringArgs
     @noKwargs
     def func_import(self, node, args, kwargs):
@@ -1018,9 +1116,33 @@ class Interpreter():
         if not isinstance(sources, list):
             sources = [sources]
         sources = self.source_strings_to_files(self.flatten(sources))
-        dep = dependencies.InternalDependency(incs, libs, sources)
+        deps = kwargs.get('dependencies', [])
+        if not isinstance(deps, list):
+            deps = [deps]
+        final_deps = []
+        for d in deps:
+            try:
+                d = d.held_object
+            except Exception:
+                pass
+            if not isinstance(d, (dependencies.Dependency, dependencies.ExternalLibrary)):
+                raise InterpreterException('Dependencies must be external deps')
+            final_deps.append(d)
+        dep = dependencies.InternalDependency(incs, libs, sources, final_deps)
         return InternalDependencyHolder(dep)
 
+    @noKwargs
+    def func_assert(self, node, args, kwargs):
+        if len(args) != 2:
+            raise InterpreterException('Assert takes exactly two arguments')
+        value, message = args
+        if not isinstance(value, bool):
+            raise InterpreterException('Assert value not bool.')
+        if not isinstance(message, str):
+            raise InterpreterException('Assert message not a string.')
+        if not value:
+            raise InterpreterException('Assert failed: ' + message)
+
     def set_variable(self, varname, variable):
         if variable is None:
             raise InvalidCode('Can not assign None to variable.')
@@ -1153,11 +1275,13 @@ class Interpreter():
         self.build.pkgconfig_gens.append(p)
 
     @stringArgs
-    @noKwargs
     def func_subproject(self, nodes, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Subproject takes exactly one argument')
         dirname = args[0]
+        return self.do_subproject(dirname, kwargs)
+
+    def do_subproject(self, dirname, kwargs):
         if self.subdir != '':
             segs = os.path.split(self.subdir)
             if len(segs) != 2 or segs[0] != self.subproject_dir:
@@ -1176,12 +1300,17 @@ class Interpreter():
         os.makedirs(os.path.join(self.build.environment.get_build_dir(), subdir), exist_ok=True)
         self.global_args_frozen = True
         mlog.log('\nExecuting subproject ', mlog.bold(dirname), '.\n', sep='')
-        subi = Interpreter(self.build, dirname, subdir, self.subproject_dir)
+        subi = Interpreter(self.build, self.backend, dirname, subdir, self.subproject_dir)
         subi.subprojects = self.subprojects
 
         subi.subproject_stack = self.subproject_stack + [dirname]
         current_active = self.active_projectname
         subi.run()
+        if 'version' in kwargs:
+            pv = subi.project_version
+            wanted = kwargs['version']
+            if not mesonlib.version_compare(pv, wanted):
+                raise InterpreterException('Subproject %s version is %s but %s required.' % (dirname, pv, wanted))
         self.active_projectname = current_active
         mlog.log('\nSubproject', mlog.bold(dirname), 'finished.')
         self.build.subprojects[dirname] = True
@@ -1196,15 +1325,20 @@ class Interpreter():
         if len(args) != 1:
             raise InterpreterException('Argument required for get_option.')
         optname = args[0]
-        if optname not in coredata.builtin_options and self.is_subproject():
-            optname = self.subproject + ':' + optname
         try:
             return self.environment.get_coredata().get_builtin_option(optname)
         except RuntimeError:
             pass
-        if optname not in self.environment.coredata.user_options:
+        try:
+            return self.environment.coredata.compiler_options[optname].value
+        except KeyError:
+            pass
+        if optname not in coredata.builtin_options and self.is_subproject():
+            optname = self.subproject + ':' + optname
+        try:
+            return self.environment.coredata.user_options[optname].value
+        except KeyError:
             raise InterpreterException('Tried to access unknown option "%s".' % optname)
-        return self.environment.coredata.user_options[optname].value
 
     @noKwargs
     def func_configuration_data(self, node, args, kwargs):
@@ -1212,6 +1346,26 @@ class Interpreter():
             raise InterpreterException('configuration_data takes no arguments')
         return ConfigurationDataHolder()
 
+    def parse_default_options(self, default_options):
+        if not isinstance(default_options, list):
+            default_options = [default_options]
+        for option in default_options:
+            if not isinstance(option, str):
+                mlog.debug(option)
+                raise InterpreterException('Default options must be strings')
+            if '=' not in option:
+                raise InterpreterException('All default options must be of type key=value.')
+            key, value = option.split('=', 1)
+            builtin_options = self.coredata.builtin_options
+            if key in builtin_options:
+                if not hasattr(self.environment.cmd_line_options, value):
+                    self.coredata.set_builtin_option(key, value)
+                # If this was set on the command line, do not override.
+            else:
+                newoptions = [option] + self.environment.cmd_line_options.projectoptions
+                self.environment.cmd_line_options.projectoptions = newoptions
+
+
     @stringArgs
     def func_project(self, node, args, kwargs):
         if len(args) < 2:
@@ -1219,13 +1373,21 @@ class Interpreter():
 
         if not self.is_subproject():
             self.build.project_name = args[0]
+            if self.environment.first_invocation and 'default_options' in kwargs:
+                self.parse_default_options(kwargs['default_options'])
         self.active_projectname = args[0]
-        self.build.dep_manifest[args[0]] = kwargs.get('version', 'undefined')
+        self.project_version = kwargs.get('version', 'undefined')
+        self.build.dep_manifest[args[0]] = self.project_version
         if self.subproject in self.build.projects:
             raise InvalidCode('Second call to project().')
         if not self.is_subproject() and 'subproject_dir' in kwargs:
             self.subproject_dir = kwargs['subproject_dir']
 
+        if 'meson_version' in kwargs:
+            cv = coredata.version
+            pv = kwargs['meson_version']
+            if not mesonlib.version_compare(cv, pv):
+                raise InterpreterException('Meson version is %s but project requires %s.' % (cv, pv))
         self.build.projects[self.subproject] = args[0]
         mlog.log('Project name: ', mlog.bold(args[0]), sep='')
         self.add_languages(node, args[1:])
@@ -1242,7 +1404,7 @@ class Interpreter():
     @noKwargs
     def func_message(self, node, args, kwargs):
         # reduce arguments again to avoid flattening posargs
-        (posargs, kwargs) = self.reduce_arguments(node.args)
+        (posargs, _) = self.reduce_arguments(node.args)
         if len(posargs) != 1:
             raise InvalidArguments('Expected 1 argument, got %d' % len(posargs))
 
@@ -1317,6 +1479,18 @@ class Interpreter():
                 if cross_comp is not None:
                     cross_comp.sanity_check(self.environment.get_scratch_dir())
                     self.coredata.cross_compilers[lang] = cross_comp
+                new_options = comp.get_options()
+                optprefix = lang + '_'
+                for i in new_options:
+                    if not i.startswith(optprefix):
+                        raise InterpreterException('Internal error, %s has incorrect prefix.' % i)
+                    cmd_prefix = i + '='
+                    for cmd_arg in self.environment.cmd_line_options.projectoptions:
+                        if cmd_arg.startswith(cmd_prefix):
+                            value = cmd_arg.split('=', 1)[1]
+                            new_options[i].value = value
+                new_options.update(self.coredata.compiler_options)
+                self.coredata.compiler_options = new_options
             mlog.log('Native %s compiler: ' % lang, mlog.bold(' '.join(comp.get_exelist())), ' (%s %s)' % (comp.id, comp.version), sep='')
             if not comp.get_language() in self.coredata.external_args:
                 (ext_compile_args, ext_link_args) = environment.get_args_from_envvars(comp.get_language())
@@ -1353,9 +1527,14 @@ class Interpreter():
         if not isinstance(required, bool):
             raise InvalidArguments('"required" argument must be a boolean.')
         libname = args[0]
-        if libname in self.coredata.ext_libs and\
-           self.coredata.ext_libs[libname].found():
-            return ExternalLibraryHolder(self.coredata.ext_libs[libname])
+        # We do not cache found libraries because they can come
+        # and go between invocations wildly. As an example we
+        # may find the 64 bit version but need instead the 32 bit
+        # one that is not installed. If we cache the found path
+        # then we will never found the new one if it get installed.
+        # This causes a bit of a slowdown as libraries are rechecked
+        # on every regen, but since it is a fast operation it should be
+        # ok.
         if 'dirs' in kwargs:
             search_dirs = kwargs['dirs']
             if not isinstance(search_dirs, list):
@@ -1370,7 +1549,6 @@ class Interpreter():
         result = self.environment.find_library(libname, search_dirs)
         extlib = dependencies.ExternalLibrary(libname, result)
         libobj = ExternalLibraryHolder(extlib)
-        self.coredata.ext_libs[libname] = extlib
         if required and not libobj.found():
             raise InvalidArguments('External library "%s" not found.' % libname)
         return libobj
@@ -1384,10 +1562,24 @@ class Interpreter():
         else:
             dep = dependencies.Dependency() # Returns always false for dep.found()
         if not dep.found():
-            dep = dependencies.find_external_dependency(name, self.environment, kwargs)
+            try:
+                dep = dependencies.find_external_dependency(name, self.environment, kwargs)
+            except dependencies.DependencyException:
+                if 'fallback' in kwargs:
+                    return self.dependency_fallback(kwargs)
+                raise
         self.coredata.deps[identifier] = dep
         return DependencyHolder(dep)
 
+    def dependency_fallback(self, kwargs):
+        fbinfo = kwargs['fallback']
+        check_stringlist(fbinfo)
+        if len(fbinfo) != 2:
+            raise InterpreterException('Fallback info must have exactly two items.')
+        dirname, varname = fbinfo
+        self.do_subproject(dirname, kwargs)
+        return self.subprojects[dirname].get_variable_method([varname], {})
+
     def func_executable(self, node, args, kwargs):
         return self.build_target(node, args, kwargs, ExecutableHolder)
 
@@ -1397,6 +1589,11 @@ class Interpreter():
     def func_shared_lib(self, node, args, kwargs):
         return self.build_target(node, args, kwargs, SharedLibraryHolder)
 
+    def func_library(self, node, args, kwargs):
+        if self.coredata.get_builtin_option('default_library') == 'shared':
+            return self.func_shared_lib(node, args, kwargs)
+        return self.func_static_lib(node, args, kwargs)
+
     def func_jar(self, node, args, kwargs):
         return self.build_target(node, args, kwargs, JarHolder)
 
@@ -1410,6 +1607,8 @@ class Interpreter():
             return self.func_shared_lib(node, args, kwargs)
         elif target_type == 'static_library':
             return self.func_static_lib(node, args, kwargs)
+        elif target_type == 'library':
+            return self.func_library(node, args, kwargs)
         elif target_type == 'jar':
             return self.func_jar(node, args, kwargs)
         else:
@@ -1455,17 +1654,19 @@ class Interpreter():
     def func_run_target(self, node, args, kwargs):
         if len(args) < 2:
             raise InterpreterException('Incorrect number of arguments')
+        cleaned_args = []
         for i in args:
             try:
                 i = i.held_object
             except AttributeError:
                 pass
-            if not isinstance(i, (str, build.BuildTarget)):
+            if not isinstance(i, (str, build.BuildTarget, build.CustomTarget)):
                 mlog.debug('Wrong type:', str(i))
                 raise InterpreterException('Invalid argument to run_target.')
-        name = args[0]
-        command = args[1]
-        cmd_args = args[2:]
+            cleaned_args.append(i)
+        name = cleaned_args[0]
+        command = cleaned_args[1]
+        cmd_args = cleaned_args[2:]
         tg = RunTargetHolder(name, command, cmd_args, self.subdir)
         self.add_target(name, tg.held_object)
         return tg
@@ -1629,7 +1830,7 @@ class Interpreter():
             absdir = os.path.join(absbase, a)
             if not os.path.isdir(absdir):
                 raise InvalidArguments('Include dir %s does not exist.' % a)
-        i = IncludeDirsHolder(self.subdir, args)
+        i = IncludeDirsHolder(build.IncludeDirs(self.subdir, args))
         return i
 
     @stringArgs
@@ -1727,7 +1928,7 @@ class Interpreter():
             mlog.debug('Unknown target type:', str(targetholder))
             raise RuntimeError('Unreachable code')
         target = targetclass(name, self.subdir, self.subproject, is_cross, sources, objs, self.environment, kwargs)
-        l = targetholder(target)
+        l = targetholder(target, self)
         self.add_target(name, l.held_object)
         self.global_args_frozen = True
         return l
@@ -1787,12 +1988,12 @@ class Interpreter():
 
     def string_method_call(self, obj, method_name, args):
         obj = self.to_native(obj)
+        (posargs, _) = self.reduce_arguments(args)
         if method_name == 'strip':
             return obj.strip()
         elif method_name == 'format':
             return self.format_string(obj, args)
         elif method_name == 'split':
-            (posargs, _) = self.reduce_arguments(args)
             if len(posargs) > 1:
                 raise InterpreterException('Split()  must have at most one argument.')
             elif len(posargs) == 1:
@@ -1802,6 +2003,13 @@ class Interpreter():
                 return obj.split(s)
             else:
                 return obj.split()
+        elif method_name == 'startswith' or method_name == 'endswith':
+            s = posargs[0]
+            if not isinstance(s, str):
+                raise InterpreterException('Argument must be a string.')
+            if method_name == 'startswith':
+                return obj.startswith(s)
+            return obj.endswith(s)
         raise InterpreterException('Unknown method "%s" for a string.' % method_name)
 
     def to_native(self, arg):
@@ -1830,7 +2038,7 @@ class Interpreter():
         else:
             obj = self.evaluate_statement(invokable)
         method_name = node.name
-        if method_name == 'extract_objects' and self.environment.coredata.unity:
+        if method_name == 'extract_objects' and self.environment.coredata.get_builtin_option('unity'):
             raise InterpreterException('Single object files can not be extracted in Unity builds.')
         args = node.args
         if isinstance(obj, mparser.StringNode):
@@ -1844,7 +2052,7 @@ class Interpreter():
         (args, kwargs) = self.reduce_arguments(args)
         if method_name == 'extract_objects':
             self.validate_extraction(obj.held_object)
-        return obj.method_call(method_name, args, kwargs)
+        return obj.method_call(method_name, self.flatten(args), kwargs)
 
     # Only permit object extraction from the same subproject
     def validate_extraction(self, buildtarget):
