@@ -22,21 +22,13 @@ import build
 import platform
 import mlog, coredata
 
-from coredata import MesonException
+from coredata import MesonException, build_types, layouts, warning_levels, libtypelist
+
+backendlist = ['ninja', 'vs2010', 'xcode']
 
 parser = argparse.ArgumentParser()
 
-backendlist = ['ninja', 'vs2010', 'xcode']
-build_types = ['plain', 'debug', 'debugoptimized', 'release']
-warning_levels = ['1', '2', '3']
-
 default_warning = '1'
-try:
-    warn_candidate = os.environ['MESON_WARN_LEVEL']
-    if warn_candidate in warning_levels:
-        default_warning = warn_candidate
-except KeyError:
-    pass
 
 if mesonlib.is_windows():
     def_prefix = 'c:/'
@@ -71,13 +63,17 @@ parser.add_argument('--unity', action='store_true', dest='unity', default=False,
                     help='unity build')
 parser.add_argument('--werror', action='store_true', dest='werror', default=False,\
                     help='Treat warnings as errors')
+parser.add_argument('--layout', choices=layouts, dest='layout', default='mirror',\
+                    help='Build directory layout.')
+parser.add_argument('--default-library', choices=libtypelist, dest='default_library',
+                    default='shared', help='Default library type.')
 parser.add_argument('--warnlevel', default=default_warning, dest='warning_level', choices=warning_levels,\
                     help='Level of compiler warnings to use (larger is more, default is %(default)s)')
 parser.add_argument('--cross-file', default=None, dest='cross_file',
                     help='file describing cross compilation environment')
 parser.add_argument('-D', action='append', dest='projectoptions', default=[],
                     help='Set project options.')
-parser.add_argument('-v', action='store_true', dest='print_version', default=False,
+parser.add_argument('-v', '--version', action='store_true', dest='print_version', default=False,
                     help='Print version.')
 parser.add_argument('directories', nargs='*')
 
@@ -141,24 +137,28 @@ itself as required.'''
         else:
             mlog.log('Build type:', mlog.bold('native build'))
         b = build.Build(env)
-        intr = interpreter.Interpreter(b)
-        mlog.log('Build machine cpu:', mlog.bold(intr.builtin['build_machine'].cpu_method([], {})))
-        if env.is_cross_build():
-            mlog.log('Host machine cpu:', mlog.bold(intr.builtin['host_machine'].cpu_method([], {})))
-            mlog.log('Target machine cpu:', mlog.bold(intr.builtin['target_machine'].cpu_method([], {})))
-        intr.run()
         if self.options.backend == 'ninja':
             import ninjabackend
-            g = ninjabackend.NinjaBackend(b, intr)
+            g = ninjabackend.NinjaBackend(b)
         elif self.options.backend == 'vs2010':
             import vs2010backend
-            g = vs2010backend.Vs2010Backend(b, intr)
+            g = vs2010backend.Vs2010Backend(b)
         elif self.options.backend == 'xcode':
             import xcodebackend
-            g = xcodebackend.XCodeBackend(b, intr)
+            g = xcodebackend.XCodeBackend(b)
         else:
             raise RuntimeError('Unknown backend "%s".' % self.options.backend)
-        g.generate()
+
+        intr = interpreter.Interpreter(b, g)
+        if env.is_cross_build():
+            mlog.log('Host machine cpu family:', mlog.bold(intr.builtin['host_machine'].cpu_family_method([], {})))
+            mlog.log('Host machine cpu:', mlog.bold(intr.builtin['host_machine'].cpu_method([], {})))
+            mlog.log('Target machine cpu family:', mlog.bold(intr.builtin['target_machine'].cpu_family_method([], {})))
+            mlog.log('Target machine cpu:', mlog.bold(intr.builtin['target_machine'].cpu_method([], {})))
+        mlog.log('Build machine cpu family:', mlog.bold(intr.builtin['build_machine'].cpu_family_method([], {})))
+        mlog.log('Build machine cpu:', mlog.bold(intr.builtin['build_machine'].cpu_method([], {})))
+        intr.run()
+        g.generate(intr)
         env.generating_finished()
         dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
         pickle.dump(b, open(dumpfile, 'wb'))
