@@ -154,7 +154,7 @@ def default_libdir():
     try:
         archpath = subprocess.check_output(['dpkg-architecture', '-qDEB_HOST_MULTIARCH']).decode().strip()
         return 'lib/' + archpath
-    except FileNotFoundError:
+    except:
         pass
     if os.path.isdir('/usr/lib64'):
         return 'lib64'
@@ -255,3 +255,12 @@ def replace_if_different(dst, dst_tmp):
         pass
     os.replace(dst_tmp, dst)
 
+def stringlistify(item):
+    if isinstance(item, str):
+        item = [item]
+    if not isinstance(item, list):
+        raise MesonException('Item is not an array')
+    for i in item:
+        if not isinstance(i, str):
+            raise MesonException('List item not a string.')
+    return item
