@@ -20,6 +20,8 @@ import os, sys
 import subprocess
 from coredata import MesonException
 import mlog
+import xml.etree.ElementTree as ET
+from mesonlib import File
 
 girwarning_printed = False
 
@@ -28,21 +30,54 @@ class GnomeModule:
     def compile_resources(self, state, args, kwargs):
         cmd = ['glib-compile-resources', '@INPUT@', '--generate']
         if 'source_dir' in kwargs:
-            d = os.path.join(state.build_to_src, state.subdir, kwargs.pop('source_dir'))
+            resource_loc = os.path.join(state.subdir, kwargs.pop('source_dir'))
+            d = os.path.join(state.build_to_src, resource_loc)
             cmd += ['--sourcedir', d]
+        else:
+            resource_loc = state.subdir
         if 'c_name' in kwargs:
             cmd += ['--c-name', kwargs.pop('c_name')]
         cmd += ['--target', '@OUTPUT@']
         kwargs['command'] = cmd
         output_c = args[0] + '.c'
         output_h = args[0] + '.h'
-        kwargs['input'] = args[1]
+        resfile = args[1]
+        kwargs['depend_files'] = self.parse_gresource_xml(state, resfile, resource_loc)
+        kwargs['input'] = resfile
         kwargs['output'] = output_c
         target_c = build.CustomTarget(args[0]+'_c', state.subdir, kwargs)
         kwargs['output'] = output_h
         target_h = build.CustomTarget(args[0] + '_h', state.subdir, kwargs)
         return [target_c, target_h]
-    
+
+    def parse_gresource_xml(self, state, fobj, resource_loc):
+        if isinstance(fobj, File):
+            fname = fobj.fname
+            subdir = fobj.subdir
+        else:
+            fname = fobj
+            subdir = state.subdir
+        abspath = os.path.join(state.environment.source_dir, state.subdir, fname)
+        relative_part = os.path.split(fname)[0]
+        try:
+            tree = ET.parse(abspath)
+            root = tree.getroot()
+            result = []
+            for child in root[0]:
+                if child.tag != 'file':
+                    mlog.log("Warning, malformed rcc file: ", os.path.join(state.subdir, fname))
+                    break
+                else:
+                    relfname = os.path.join(resource_loc, child.text)
+                    absfname = os.path.join(state.environment.source_dir, relfname)
+                    if os.path.isfile(absfname):
+                        result.append(relfname)
+                    else:
+                        mlog.log('Warning, resource file points to nonexisting file %s.' % relfname)
+            return result
+        except Exception:
+            return []
+
     def generate_gir(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gir takes one argument')
@@ -139,15 +174,15 @@ class GnomeModule:
             inc_dirs = kwargs.pop('include_directories')
             if not isinstance(inc_dirs, list):
                 inc_dirs = [inc_dirs]
-            for id in inc_dirs:
-                if isinstance(id.held_object, build.IncludeDirs):
-                    scan_command += ['--add-include-path=%s' % inc for inc in id.held_object.get_incdirs()]
+            for ind in inc_dirs:
+                if isinstance(ind.held_object, build.IncludeDirs):
+                    scan_command += ['--add-include-path=%s' % inc for inc in ind.held_object.get_incdirs()]
                 else:
                     raise MesonException('Gir include dirs should be include_directories()')
         if isinstance(girtarget, build.Executable):
             scan_command += ['--program', girtarget]
         elif isinstance(girtarget, build.SharedLibrary):
-            scan_command += ["-L", os.path.join (state.environment.get_build_dir(), girtarget.subdir)]
+            scan_command += ["-L@PRIVATE_OUTDIR_ABS_%s@" % girtarget.get_id()]
             libname = girtarget.get_basename()
             scan_command += ['--library', libname]
         scankwargs = {'output' : girfile,
@@ -163,9 +198,9 @@ class GnomeModule:
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
         typelib_cmd = ['g-ir-compiler', scan_target, '--output', '@OUTPUT@']
         if inc_dirs:
-            for id in inc_dirs:
+            for incd in inc_dirs:
                 typelib_cmd += ['--includedir=%s' % inc for inc in
-                                id.held_object.get_incdirs()]
+                                incd.held_object.get_incdirs()]
         if deps:
             for dep in deps:
                 girdir = dep.held_object.get_variable ("girdir")
@@ -227,17 +262,33 @@ class GnomeModule:
             header_dir = os.path.join(state.environment.get_source_dir(), src_dir.get_curdir(), incdirs[0])
         else:
             header_dir = os.path.normpath(os.path.join(state.subdir, src_dir))
-        args = [state.environment.get_source_dir(),
-                state.environment.get_build_dir(),
-                state.subdir,
-                header_dir,
-                main_file,
-                modulename]
+        args = ['--sourcedir=' + state.environment.get_source_dir(),
+                '--builddir=' + state.environment.get_build_dir(),
+                '--subdir=' + state.subdir,
+                '--headerdir=' + header_dir,
+                '--mainfile=' + main_file,
+                '--modulename=' + modulename]
+        args += self.unpack_args('--htmlargs=', 'html_args', kwargs)
+        args += self.unpack_args('--scanargs=', 'scan_args', kwargs)
         res = [build.RunTarget(targetname, command, args, state.subdir)]
         if kwargs.get('install', True):
             res.append(build.InstallScript([command] + args))
         return res
 
+    def unpack_args(self, arg, kwarg_name, kwargs):
+        try:
+            new_args = kwargs[kwarg_name]
+            if not isinstance(new_args, list):
+                new_args = [new_args]
+            for i in new_args:
+                if not isinstance(i, str):
+                    raise MesonException('html_args values must be strings.')
+        except KeyError:
+            return[]
+        if len(new_args) > 0:
+            return [arg + '@@'.join(new_args)]
+        return []
+
     def gdbus_codegen(self, state, args, kwargs):
         if len(args) != 2:
             raise MesonException('Gdbus_codegen takes two arguments, name and xml file.')
@@ -248,7 +299,7 @@ class GnomeModule:
             cmd += ['--interface-prefix', kwargs.pop('interface_prefix')]
         if 'namespace' in kwargs:
             cmd += ['--c-namespace', kwargs.pop('namespace')]
-        cmd += ['--generate-c-code', os.path.join(state.subdir, namebase), '@INPUT@']
+        cmd += ['--generate-c-code', '@OUTDIR@/' + namebase, '@INPUT@']
         outputs = [namebase + '.c', namebase + '.h']
         custom_kwargs = {'input' : xml_file,
                          'output' : outputs,
