@@ -53,6 +53,7 @@ class Lexer:
             ('dblquote', re.compile(r'"')),
             ('string', re.compile(r"'([^'\\]|(\\.))*'")),
             ('comma', re.compile(r',')),
+            ('plusassign', re.compile(r'\+=')),
             ('dot', re.compile(r'\.')),
             ('plus', re.compile(r'\+')),
             ('dash', re.compile(r'-')),
@@ -209,6 +210,13 @@ class CodeBlockNode:
         self.colno = colno
         self.lines = []
 
+class IndexNode:
+    def __init__(self, iobject, index):
+        self.iobject = iobject
+        self.index = index
+        self.lineno = iobject.lineno
+        self.colno = iobject.colno
+
 class MethodNode:
     def __init__(self, lineno, colno, source_object, name, args):
         self.lineno = lineno
@@ -234,6 +242,14 @@ class AssignmentNode:
         assert(isinstance(var_name, str))
         self.value = value
 
+class PlusAssignmentNode:
+    def __init__(self, lineno, colno, var_name, value):
+        self.lineno = lineno
+        self.colno = colno
+        self.var_name = var_name
+        assert(isinstance(var_name, str))
+        self.value = value
+
 class ForeachClauseNode():
     def __init__(self, lineno, colno, varname, items, block):
         self.lineno = lineno
@@ -344,7 +360,12 @@ class Parser:
 
     def e1(self):
         left = self.e2()
-        if self.accept('assign'):
+        if self.accept('plusassign'):
+            value = self.e1()
+            if not isinstance(left, IdNode):
+                raise ParseException('Plusassignment target must be an id.', left.lineno, left.colno)
+            return PlusAssignmentNode(left.lineno, left.colno, left.value, value)
+        elif self.accept('assign'):
             value = self.e1()
             if not isinstance(left, IdNode):
                 raise ParseException('Assignment target must be an id.',
@@ -415,8 +436,15 @@ class Parser:
                 raise ParseException('Function call must be applied to plain id',
                                      left.lineno, left.colno)
             left = FunctionNode(left.lineno, left.colno, left.value, args)
-        while self.accept('dot'):
-            left = self.method_call(left)
+        go_again = True
+        while go_again:
+            go_again = False
+            if self.accept('dot'):
+                go_again = True
+                left = self.method_call(left)
+            if self.accept('lbracket'):
+                go_again = True
+                left = self.index_call(left)
         return left
 
     def e8(self):
@@ -478,6 +506,11 @@ class Parser:
             return self.method_call(method)
         return method
 
+    def index_call(self, source_object):
+        index_statement = self.statement()
+        self.expect('rbracket')
+        return IndexNode(source_object, index_statement)
+
     def foreachblock(self):
         t = self.current
         self.expect('id')
