@@ -21,7 +21,7 @@ from mesonlib import File
 from meson_install import InstallData
 from build import InvalidArguments
 from coredata import MesonException
-import os, sys, shutil, pickle, re
+import os, sys, pickle, re
 
 if mesonlib.is_windows():
     quote_char = '"'
@@ -85,6 +85,12 @@ class NinjaBuildElement():
         if len(self.orderdeps) > 0:
             line += ' || ' + ' '.join([ninja_quote(x) for x in self.orderdeps])
         line += '\n'
+        # This needs to be done to make these strings
+        # pass through arbitrary shells. Backslash is a
+        # quote character so it can break at any time.
+        # Because of this always use forward slashes,
+        # it is a path separator even on Windows.
+        line = line.replace('\\', '/')
         outfile.write(line)
 
         for e in self.elems:
@@ -104,13 +110,14 @@ class NinjaBuildElement():
                 newelems.append(templ % ninja_quote(i))
             line += ' '.join(newelems)
             line += '\n'
+            line = line.replace('\\', '/')
             outfile.write(line)
         outfile.write('\n')
 
 class NinjaBackend(backends.Backend):
 
-    def __init__(self, build, interp):
-        super().__init__(build, interp)
+    def __init__(self, build):
+        super().__init__(build)
         self.source_suffix_in_objs = True
         self.ninja_filename = 'build.ninja'
         self.fortran_deps = {}
@@ -122,7 +129,8 @@ class NinjaBackend(backends.Backend):
                 raise MesonException('Multiple producers for Ninja target "%s". Please rename your targets.' % n)
             self.all_outputs[n] = True
 
-    def generate(self):
+    def generate(self, interp):
+        self.interpreter = interp
         outfilename = os.path.join(self.environment.get_build_dir(), self.ninja_filename)
         tempfilename = outfilename + '~'
         outfile = open(tempfilename, 'w')
@@ -142,7 +150,7 @@ class NinjaBackend(backends.Backend):
         self.generate_tests(outfile)
         outfile.write('# Install rules\n\n')
         self.generate_install(outfile)
-        if self.environment.coredata.coverage:
+        if self.environment.coredata.get_builtin_option('coverage'):
             outfile.write('# Coverage rules\n\n')
             self.generate_coverage_rules(outfile)
         outfile.write('# Suffix\n\n')
@@ -161,7 +169,10 @@ class NinjaBackend(backends.Backend):
                 continue
             for src in gensource.get_outfilelist():
                 if self.environment.is_header(src):
-                    header_deps.append(src)
+                    header_deps.append(os.path.join(self.get_target_private_dir(target), src))
+        for dep in target.link_targets:
+            if isinstance(dep, (build.StaticLibrary, build.SharedLibrary)):
+                header_deps += self.get_generated_headers(dep)
         return header_deps
 
     def generate_target(self, target, outfile):
@@ -192,8 +203,8 @@ class NinjaBackend(backends.Backend):
         self.generate_custom_generator_rules(target, outfile)
         outname = self.get_target_filename(target)
         obj_list = []
-        use_pch = self.environment.coredata.use_pch
-        is_unity = self.environment.coredata.unity
+        use_pch = self.environment.coredata.get_builtin_option('use_pch')
+        is_unity = self.environment.coredata.get_builtin_option('unity')
         if use_pch and target.has_pch():
             pch_objects = self.generate_pch(target, outfile)
         else:
@@ -205,7 +216,7 @@ class NinjaBackend(backends.Backend):
         for gensource in target.get_generated_sources():
             if isinstance(gensource, build.CustomTarget):
                 for src in gensource.output:
-                    src = os.path.join(gensource.subdir, src)
+                    src = os.path.join(self.get_target_dir(gensource), src)
                     if self.environment.is_source(src) and not self.environment.is_header(src):
                         if is_unity:
                             unity_deps.append(os.path.join(self.environment.get_build_dir(), RawFilename(src)))
@@ -222,13 +233,13 @@ class NinjaBackend(backends.Backend):
             else:
                 for src in gensource.get_outfilelist():
                     if self.environment.is_object(src):
-                        obj_list.append(os.path.join(self.get_target_private_dir(target), src))
+                        obj_list.append(os.path.join(self.get_target_private_dir_abs(target), src))
                     elif not self.environment.is_header(src):
                         if is_unity:
-                            if '/' in src:
+                            if self.has_dir_part(src):
                                 rel_src = src
                             else:
-                                rel_src = os.path.join(self.get_target_private_dir(target), src)
+                                rel_src = os.path.join(self.get_target_private_dir_abs(target), src)
                             unity_deps.append(rel_src)
                             abs_src = os.path.join(self.environment.get_build_dir(), rel_src)
                             unity_src.append(abs_src)
@@ -254,8 +265,8 @@ class NinjaBackend(backends.Backend):
             if not self.environment.is_header(src):
                 src_list.append(src)
                 if is_unity:
-                    abs_src = os.path.join(self.environment.get_source_dir(),
-                                           target.get_subdir(), src)
+                    abs_src = os.path.join(self.environment.get_build_dir(),
+                                           src.rel_to_builddir(self.build_to_src))
                     unity_src.append(abs_src)
                 else:
                     obj_list.append(self.generate_single_compile(target, outfile, src, False, [], header_deps))
@@ -275,43 +286,33 @@ class NinjaBackend(backends.Backend):
             if not tname in self.processed_targets:
                 self.generate_target(t, outfile)
 
-    def hackety_hack(self, hack):
-        if isinstance(hack, list):
-            return hack[0]
-        return hack
-
     def generate_custom_target(self, target, outfile):
-        ofilenames = [os.path.join(target.subdir, i) for i in target.output]
-        # FIXME, should not grab element at zero but rather expand all.
-        deps = [os.path.join(i.get_subdir(), self.hackety_hack(i.get_filename())) for i in target.get_dependencies()]
-        srcs = []
-        for i in target.sources:
-            if isinstance(i, str):
-                srcs.append(os.path.join(self.build_to_src, target.subdir, i))
-            else:
-                srcs.append(i.rel_to_builddir(self.build_to_src))
-        deps +=  srcs
+        (srcs, ofilenames, cmd) = self.eval_custom_target_command(target)
+        deps = []
+        for i in target.get_dependencies():
+            # FIXME, should not grab element at zero but rather expand all.
+            if isinstance(i, list):
+                i = i[0]
+            fname = i.get_filename()
+            if isinstance(fname, list):
+                fname = fname[0]
+            deps.append(os.path.join(self.get_target_dir(i), fname))
         if target.build_always:
             deps.append('PHONY')
-        elem = NinjaBuildElement(ofilenames, 'CUSTOM_COMMAND', deps)
+        elem = NinjaBuildElement(ofilenames, 'CUSTOM_COMMAND', srcs)
+        for i in target.depend_files:
+            if isinstance(i, mesonlib.File):
+                deps.append(i.rel_to_builddir(self.build_to_src))
+            else:
+                deps.append(os.path.join(self.build_to_src, i))
+        elem.add_dep(deps)
         for d in target.extra_depends:
             tmp = d.get_filename()
             if not isinstance(tmp, list):
                 tmp = [tmp]
             for fname in tmp:
-                elem.add_dep(os.path.join(d.get_subdir(), fname))
-        cmd = []
-        for i in target.command:
-            for (j, src) in enumerate(srcs):
-                i = i.replace('@INPUT%d@' % j, src)
-            for (j, res) in enumerate(ofilenames):
-                i = i.replace('@OUTPUT%d@' % j, res)
-            if i == '@INPUT@':
-                cmd += srcs
-            elif i == '@OUTPUT@':
-                cmd += ofilenames
-            else:
-                cmd.append(i)
+                elem.add_dep(os.path.join(self.get_target_dir(d), fname))
+
         elem.add_item('COMMAND', cmd)
         elem.add_item('description', 'Generating %s with a custom command.' % target.name)
         elem.write(outfile)
@@ -325,10 +326,13 @@ class NinjaBackend(backends.Backend):
         for i in target.args:
             if isinstance(i, str):
                 arg_strings.append(i)
-            elif isinstance(i, build.BuildTarget):
-                deps.append(self.get_target_filename(i))
+            elif isinstance(i, (build.BuildTarget, build.CustomTarget)):
+                relfname = self.get_target_filename(i)
+                deps.append(relfname)
+                arg_strings.append(os.path.join(self.environment.get_build_dir(), relfname))
             else:
-                raise MesonException('Unreachable code.')
+                mlog.debug(str(i))
+                raise MesonException('Unreachable code in generate_run_target.')
         elem = NinjaBuildElement(target.name, 'CUSTOM_COMMAND', deps)
         cmd = [sys.executable, runnerscript, self.environment.get_source_dir(), self.environment.get_build_dir(), target.subdir]
         texe = target.command
@@ -345,7 +349,7 @@ class NinjaBackend(backends.Backend):
             cmd.append(abs_exe)
         else:
             cmd.append(target.command)
-        cmd += target.args
+        cmd += arg_strings
         elem.add_item('COMMAND', cmd)
         elem.add_item('description', 'Running external command %s.' % target.name)
         elem.add_item('pool', 'console')
@@ -443,14 +447,14 @@ class NinjaBackend(backends.Backend):
             for lang in languages:
                 rel_src =  os.path.join(subdir, lang + '.gmo')
                 src_file = os.path.join(self.environment.get_build_dir(), rel_src)
-                d.po.append((src_file, self.environment.coredata.localedir, lang))
+                d.po.append((src_file, self.environment.coredata.get_builtin_option('localedir'), lang))
                 elem.add_dep(rel_src)
 
     def generate_target_install(self, d):
         libdir = self.environment.get_libdir()
         bindir = self.environment.get_bindir()
 
-        should_strip = self.environment.coredata.strip
+        should_strip = self.environment.coredata.get_builtin_option('strip')
         for t in self.build.get_targets().values():
             if t.should_install():
                 outdir = t.get_custom_install_dir()
@@ -464,7 +468,7 @@ class NinjaBackend(backends.Backend):
                 d.targets.append(i)
 
     def generate_pkgconfig_install(self, d):
-        pkgroot = os.path.join(self.environment.coredata.libdir, 'pkgconfig')
+        pkgroot = os.path.join(self.environment.coredata.get_builtin_option('libdir'), 'pkgconfig')
 
         for p in self.build.pkgconfig_gens:
             pcfile = p.filebase + '.pc'
@@ -505,7 +509,6 @@ class NinjaBackend(backends.Backend):
                 d.man.append(i)
 
     def generate_data_install(self, d):
-        dataroot = self.environment.get_datadir()
         data = self.build.get_data()
         for de in data:
             subdir = de.get_install_dir()
@@ -597,7 +600,7 @@ class NinjaBackend(backends.Backend):
     def generate_jar_target(self, target, outfile):
         fname = target.get_filename()
         subdir = target.get_subdir()
-        outname_rel = os.path.join(subdir, fname)
+        outname_rel = os.path.join(self.get_target_dir(target), fname)
         src_list = target.get_sources()
         class_list = []
         compiler = self.get_compiler_for_source(src_list[0])
@@ -610,13 +613,13 @@ class NinjaBackend(backends.Backend):
         if main_class != '':
             e = 'e'
         for src in src_list:
-            class_list.append(self.generate_single_java_compile(subdir, src, target, compiler, outfile))
+            class_list.append(self.generate_single_java_compile(src, target, compiler, outfile))
         jar_rule = 'java_LINKER'
         commands = [c+m+e+f]
         if e != '':
             commands.append(main_class)
         commands.append(self.get_target_filename(target))
-        commands += ['-C', self.get_target_private_dir(target)]
+        commands += ['-C', self.get_target_private_dir_abs(target)]
         commands += class_list
         elem = NinjaBuildElement(outname_rel, jar_rule, [])
         elem.add_dep([os.path.join(self.get_target_private_dir(target), i) for i in class_list])
@@ -647,10 +650,9 @@ class NinjaBackend(backends.Backend):
         return (args, deps)
 
     def generate_cs_target(self, target, outfile):
-        buildtype = self.environment.coredata.buildtype
+        buildtype = self.environment.coredata.get_builtin_option('buildtype')
         fname = target.get_filename()
-        subdir = target.get_subdir()
-        outname_rel = os.path.join(subdir, fname)
+        outname_rel = os.path.join(self.get_target_dir(target), fname)
         src_list = target.get_sources()
         compiler = self.get_compiler_for_source(src_list[0])
         assert(compiler.get_language() == 'cs')
@@ -669,8 +671,9 @@ class NinjaBackend(backends.Backend):
         deps += resource_deps
         commands += compiler.get_output_args(outname_rel)
         for l in target.link_targets:
-            commands += compiler.get_link_args(l.get_filename())
-            deps.append(l.get_filename())
+            lname = os.path.join(self.get_target_dir(l), l.get_filename())
+            commands += compiler.get_link_args(lname)
+            deps.append(lname)
         if '-g' in commands:
             outputs = [outname_rel, outname_rel + '.mdb']
         else:
@@ -681,13 +684,13 @@ class NinjaBackend(backends.Backend):
         self.check_outputs(elem)
         elem.write(outfile)
 
-    def generate_single_java_compile(self, subdir, src, target, compiler, outfile):
+    def generate_single_java_compile(self, src, target, compiler, outfile):
         args = []
-        args += compiler.get_buildtype_args(self.environment.coredata.buildtype)
+        args += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
         args += compiler.get_output_args(self.get_target_private_dir(target))
         rel_src = src.rel_to_builddir(self.build_to_src)
         plain_class_path = src.fname[:-4] + 'class'
-        rel_obj = os.path.join(self.get_target_private_dir(target), plain_class_path)
+        rel_obj = os.path.join(self.get_target_private_dir_abs(target), plain_class_path)
         element = NinjaBuildElement(rel_obj, compiler.get_language() + '_COMPILER', rel_src)
         element.add_item('ARGS', args)
         element.write(outfile)
@@ -709,7 +712,7 @@ class NinjaBackend(backends.Backend):
             if not s.endswith('.vala'):
                 continue
             vapibase = os.path.basename(s.fname)[:-4] + 'vapi'
-            rel_vapi = os.path.join(self.get_target_private_dir(target), vapibase)
+            rel_vapi = os.path.join(self.get_target_private_dir_abs(target), vapibase)
             args = ['--fast-vapi=' + rel_vapi]
             rel_s = s.rel_to_builddir(self.build_to_src)
             element = NinjaBuildElement(rel_vapi, valac.get_language() + '_COMPILER', rel_s)
@@ -740,7 +743,7 @@ class NinjaBackend(backends.Backend):
         for s in src:
             if not s.endswith('.vala'):
                 continue
-            args = ['-d', self.get_target_private_dir(target)]
+            args = ['-d', self.get_target_private_dir_abs(target)]
             sc = os.path.basename(s.fname)[:-4] + 'c'
             args += ['-C']
             vapi_order_deps = []
@@ -750,10 +753,10 @@ class NinjaBackend(backends.Backend):
                 (vapibase, rel_vapi) = vapi_info
                 args += ['--use-fast-vapi=' + rel_vapi]
                 vapi_order_deps.append(rel_vapi)
-            relsc = os.path.join(self.get_target_private_dir(target), sc)
+            relsc = os.path.join(self.get_target_private_dir_abs(target), sc)
             rel_s = s.rel_to_builddir(self.build_to_src)
             args += ['--deps', relsc + '.d']
-            if self.environment.coredata.werror:
+            if self.environment.coredata.get_builtin_option('werror'):
                 args += valac.get_werror_args()
             for d in target.external_deps:
                 if isinstance(d, dependencies.PkgConfigDependency):
@@ -787,22 +790,22 @@ class NinjaBackend(backends.Backend):
         for i in target.get_sources():
             if not rustc.can_compile(i):
                 raise InvalidArguments('Rust target %s contains a non-rust source file.' % target.get_basename())
-            relsrc.append(os.path.join(self.build_to_src, i))
+            relsrc.append(i.rel_to_builddir(self.build_to_src))
         target_name = os.path.join(target.subdir, target.get_filename())
         args = ['--crate-type']
         if isinstance(target, build.Executable):
             cratetype = 'bin'
         elif isinstance(target, build.SharedLibrary):
-            cratetype = 'dylib'
+            cratetype = 'rlib'
         elif isinstance(target, build.StaticLibrary):
-            cratetype = 'lib'
+            cratetype = 'rlib'
         else:
             raise InvalidArguments('Unknown target type for rustc.')
         args.append(cratetype)
-        args += rustc.get_buildtype_args(self.environment.coredata.buildtype)
-        depfile = target_name + '.d'
+        args += rustc.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        depfile = target.name + '.d'
         args += ['--out-dir', target.subdir]
-        args += ['--dep-info', depfile]
+        args += ['--emit', 'dep-info', '--emit', 'link']
         orderdeps = [os.path.join(t.subdir, t.get_filename()) for t in target.link_targets]
         linkdirs = {}
         for d in target.link_targets:
@@ -836,7 +839,14 @@ class NinjaBackend(backends.Backend):
         if static_linker is None:
             return
         rule = 'rule STATIC%s_LINKER\n' % crstr
-        command = ' command = %s  $LINK_ARGS %s $in\n' % \
+        if mesonlib.is_windows():
+            command_templ = ''' command = %s @$out.rsp
+ rspfile = $out.rsp
+ rspfile_content = $LINK_ARGS %s $in
+'''
+        else:
+            command_templ = ' command = %s $LINK_ARGS %s $in\n'
+        command = command_templ %\
         (' '.join(static_linker.get_exelist()),
          ' '.join(static_linker.get_output_args('$out')))
         description = ' description = Static linking library $out\n\n'
@@ -869,7 +879,14 @@ class NinjaBackend(backends.Backend):
                     except KeyError:
                         pass
                 rule = 'rule %s%s_LINKER\n' % (langname, crstr)
-                command = ' command = %s %s $ARGS  %s $in $LINK_ARGS $aliasing\n' % \
+                if mesonlib.is_windows():
+                    command_template = ''' command = %s @$out.rsp
+ rspfile = $out.rsp
+ rspfile_content = %s $ARGS  %s $in $LINK_ARGS $aliasing
+'''
+                else:
+                    command_template = ' command = %s %s $ARGS  %s $in $LINK_ARGS $aliasing\n'
+                command = command_template % \
                 (' '.join(compiler.get_linker_exelist()),\
                  ' '.join(cross_args),\
                  ' '.join(compiler.get_linker_output_args('$out')))
@@ -931,12 +948,9 @@ class NinjaBackend(backends.Backend):
     def generate_rust_compile_rules(self, compiler, outfile):
         rule = 'rule %s_COMPILER\n' % compiler.get_language()
         invoc = ' '.join([ninja_quote(i) for i in compiler.get_exelist()])
-        command = ' command = %s %s $out $cratetype %s $ARGS $in\n' % \
-            (ninja_quote(sys.executable),
-             ninja_quote(os.path.join(os.path.split(__file__)[0], "rustrunner.py")),
-                         invoc)
+        command = ' command = %s $ARGS $in\n' % invoc
         description = ' description = Compiling Rust source $in.\n'
-        depfile = ' depfile = $out.d\n'
+        depfile = ' depfile = $targetdep\n'
 
         depstyle = ' deps = gcc\n'
         outfile.write(rule)
@@ -998,7 +1012,14 @@ rule FORTRAN_DEP_HACK
                 cross_args = self.environment.cross_info.config['properties'][langname + '_args']
             except KeyError:
                 pass
-        command = " command = %s %s $ARGS %s %s %s $in\n" % \
+        if mesonlib.is_windows():
+            command_template = ''' command = %s @$out.rsp
+ rspfile = $out.rsp
+ rspfile_content = %s $ARGS %s %s %s $in
+'''
+        else:
+            command_template = ' command = %s %s $ARGS %s %s %s $in\n'
+        command = command_template % \
             (' '.join(compiler.get_exelist()),\
              ' '.join(cross_args),
              ' '.join(quoted_depargs),\
@@ -1097,34 +1118,27 @@ rule FORTRAN_DEP_HACK
                 continue # Customtarget has already written its output rules
             generator = genlist.get_generator()
             exe = generator.get_exe()
-            if self.environment.is_cross_build() and \
-            isinstance(exe, build.BuildTarget) and exe.is_cross:
-                if 'exe_wrapper'  not in self.environment.cross_info:
-                    s = 'Can not use target %s as a generator because it is cross-built\n'
-                    s += 'and no exe wrapper is defined. You might want to set it to native instead.'
-                    s = s % exe.name
-                    raise MesonException(s)
+            exe_arr = self.exe_object_to_cmd_array(exe)
             infilelist = genlist.get_infilelist()
             outfilelist = genlist.get_outfilelist()
-            if isinstance(exe, build.BuildTarget):
-                exe_arr = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))]
-            else:
-                exe_arr = exe.get_command()
             base_args = generator.get_arglist()
             extra_dependencies = [os.path.join(self.build_to_src, i) for i in genlist.extra_depends]
             for i in range(len(infilelist)):
                 if len(generator.outputs) == 1:
-                    sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
+                    sole_output = os.path.join(self.get_target_private_dir_abs(target), outfilelist[i])
                 else:
                     sole_output = ''
                 curfile = infilelist[i]
                 infilename = os.path.join(self.build_to_src, curfile)
                 outfiles = genlist.get_outputs_for(curfile)
-                outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
+                outfiles = [os.path.join(self.get_target_private_dir_abs(target), of) for of in outfiles]
                 args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
                         for x in base_args]
-                args = self.replace_outputs(args, self.get_target_private_dir(target), outfilelist)
-                relout = os.path.join(target.subdir, target.get_basename() + target.type_suffix())
+                args = self.replace_outputs(args, self.get_target_private_dir_abs(target), outfilelist)
+                # We have consumed output files, so drop them from the list of remaining outputs.
+                if sole_output == '':
+                    outfilelist = outfilelist[len(generator.outputs):]
+                relout = self.get_target_private_dir(target)
                 args = [x.replace("@SOURCE_DIR@", self.build_to_src).replace("@BUILD_DIR@", relout)
                         for x in args]
                 cmdlist = exe_arr + args
@@ -1169,7 +1183,7 @@ rule FORTRAN_DEP_HACK
     def get_fortran_deps(self, compiler, src, target):
         mod_files = []
         usere = re.compile(r"\s*use\s+(\w+)", re.IGNORECASE)
-        dirname = self.get_target_private_dir(target)
+        dirname = self.get_target_private_dir_abs(target)
         tdeps= self.fortran_deps[target.get_basename()]
         for line in open(src):
             usematch = usere.match(line)
@@ -1216,10 +1230,10 @@ rule FORTRAN_DEP_HACK
         if isinstance(src, RawFilename):
             rel_src = src.fname
         elif is_generated:
-            if '/' in src:
+            if self.has_dir_part(src):
                 rel_src = src
             else:
-                rel_src = os.path.join(self.get_target_private_dir(target), src)
+                rel_src = os.path.join(self.get_target_private_dir_abs(target), src)
                 abs_src = os.path.join(self.environment.get_source_dir(), rel_src)
         else:
             if isinstance(src, File):
@@ -1239,7 +1253,7 @@ rule FORTRAN_DEP_HACK
         rel_obj = os.path.join(self.get_target_private_dir(target), obj_basename)
         rel_obj += '.' + self.environment.get_object_suffix()
         dep_file = rel_obj + '.' + compiler.get_depfile_suffix()
-        if self.environment.coredata.use_pch:
+        if self.environment.coredata.get_builtin_option('use_pch'):
             pchlist = target.get_pch(compiler.language)
         else:
             pchlist = []
@@ -1247,7 +1261,7 @@ rule FORTRAN_DEP_HACK
             pch_dep = []
         else:
             arr = []
-            i = os.path.join(self.get_target_private_dir(target), compiler.get_pch_name(pchlist[0]))
+            i = os.path.join(self.get_target_private_dir_abs(target), compiler.get_pch_name(pchlist[0]))
             arr.append(i)
             pch_dep = arr
         for i in target.get_include_dirs():
@@ -1259,7 +1273,17 @@ rule FORTRAN_DEP_HACK
                 sargs = compiler.get_include_args(srctreedir)
                 commands += bargs
                 commands += sargs
-        if self.environment.coredata.use_pch:
+            for d in i.get_extra_build_dirs():
+                commands += compiler.get_include_args(d)
+        custom_target_include_dirs = []
+        for i in target.generated:
+            if isinstance(i, build.CustomTarget):
+                idir = self.get_target_dir(i)
+                if idir not in custom_target_include_dirs:
+                    custom_target_include_dirs.append(idir)
+        for i in custom_target_include_dirs:
+            commands+= compiler.get_include_args(i)
+        if self.environment.coredata.get_builtin_option('use_pch'):
             commands += self.get_pch_include_args(compiler, target)
         crstr = ''
         if target.is_cross:
@@ -1277,22 +1301,22 @@ rule FORTRAN_DEP_HACK
                     depelem = NinjaBuildElement(modfile, 'FORTRAN_DEP_HACK', rel_obj)
                     depelem.write(outfile)
                     self.check_outputs(depelem)
-            commands += compiler.get_module_outdir_args(self.get_target_private_dir(target))
+            commands += compiler.get_module_outdir_args(self.get_target_private_dir_abs(target))
 
         element = NinjaBuildElement(rel_obj, compiler_name, rel_src)
         for d in header_deps:
             if isinstance(d, RawFilename):
                 d = d.fname
-            elif not '/' in d:
-                d = os.path.join(self.get_target_private_dir(target), d)
+            elif not self.has_dir_part(d):
+                d = os.path.join(self.get_target_private_dir_abs(target), d)
             element.add_dep(d)
         for d in extra_deps:
             element.add_dep(d)
         for d in order_deps:
             if isinstance(d, RawFilename):
                 d = d.fname
-            elif not '/' in d :
-                d = os.path.join(self.get_target_private_dir(target), d)
+            elif not self.has_dir_part(d):
+                d = os.path.join(self.get_target_private_dir_abs(target), d)
             element.add_orderdep(d)
         element.add_orderdep(pch_dep)
         element.add_orderdep(extra_orderdeps)
@@ -1304,16 +1328,19 @@ rule FORTRAN_DEP_HACK
         self.check_outputs(element)
         return rel_obj
 
+    def has_dir_part(self, fname):
+        return '/' in fname or '\\' in fname
+
     # Fortran is a bit weird (again). When you link against a library, just compiling a source file
     # requires the mod files that are output when single files are built. To do this right we would need to
     # scan all inputs and write out explicit deps for each file. That is stoo slow and too much effort so
-    # instead just have an ordered dependendy on the library. This ensures all reqquired mod files are created.
+    # instead just have an ordered dependendy on the library. This ensures all required mod files are created.
     # The real deps are then detected via dep file generation from the compiler. This breaks on compilers that
     # produce incorrect dep files but such is life.
     def get_fortran_orderdeps(self, target, compiler):
         if compiler.language != 'fortran':
             return []
-        return [os.path.join(lt.subdir, lt.filename) for lt in target.link_targets]
+        return [os.path.join(self.get_target_dir(lt), lt.get_filename()) for lt in target.link_targets]
 
     def generate_msvc_pch_command(self, target, compiler, pch):
         if len(pch) != 2:
@@ -1321,7 +1348,7 @@ rule FORTRAN_DEP_HACK
         header = pch[0]
         source = pch[1]
         pchname = compiler.get_pch_name(header)
-        dst = os.path.join(self.get_target_private_dir(target), pchname)
+        dst = os.path.join(self.get_target_private_dir_abs(target), pchname)
 
         commands = []
         commands += self.generate_basic_compiler_args(target, compiler)
@@ -1334,7 +1361,7 @@ rule FORTRAN_DEP_HACK
     def generate_gcc_pch_command(self, target, compiler, pch):
         commands = []
         commands += self.generate_basic_compiler_args(target, compiler)
-        dst = os.path.join(self.get_target_private_dir(target),
+        dst = os.path.join(self.get_target_private_dir_abs(target),
                            os.path.split(pch)[-1] + '.' + compiler.get_pch_suffix())
         dep = dst + '.' + compiler.get_depfile_suffix()
         return (commands, dep, dst, []) # Gcc does not create an object file during pch generation.
@@ -1372,7 +1399,7 @@ rule FORTRAN_DEP_HACK
 
     def generate_shsym(self, outfile, target):
         target_name = self.get_target_filename(target)
-        targetdir = self.get_target_private_dir(target)
+        targetdir = self.get_target_private_dir_abs(target)
         symname = os.path.join(targetdir, target_name + '.symbols')
         elem = NinjaBuildElement(symname, 'SHSYM', target_name)
         if self.environment.is_cross_build() and self.environment.cross_info.need_cross_compiler():
@@ -1394,7 +1421,8 @@ rule FORTRAN_DEP_HACK
         abspath = os.path.join(self.environment.get_build_dir(), target.subdir)
         commands = []
         commands += linker.get_linker_always_args()
-        commands += linker.get_buildtype_linker_args(self.environment.coredata.buildtype)
+        commands += linker.get_buildtype_linker_args(self.environment.coredata.get_builtin_option('buildtype'))
+        commands += linker.get_option_link_args(self.environment.coredata.compiler_options)
         if not(isinstance(target, build.StaticLibrary)):
             commands += self.environment.coredata.external_link_args[linker.get_language()]
         if isinstance(target, build.Executable):
@@ -1422,7 +1450,8 @@ rule FORTRAN_DEP_HACK
         for d in target.external_deps:
             if d.need_threads():
                 commands += linker.thread_link_flags()
-        commands += target.link_args
+        if not isinstance(target, build.StaticLibrary):
+            commands += target.link_args
         # External deps must be last because target link libraries may depend on them.
         if not(isinstance(target, build.StaticLibrary)):
             for dep in target.get_external_deps():
@@ -1432,8 +1461,8 @@ rule FORTRAN_DEP_HACK
                     for dep in d.get_external_deps():
                         commands += dep.get_link_args()
         commands += linker.build_rpath_args(self.environment.get_build_dir(),\
-                                            target.get_rpaths(), target.install_rpath)
-        if self.environment.coredata.coverage:
+                                            self.determine_rpath_dirs(target), target.install_rpath)
+        if self.environment.coredata.get_builtin_option('coverage'):
             commands += linker.get_coverage_link_args()
         commands += extra_args
         dep_targets = [self.get_dependency_filename(t) for t in dependencies]
@@ -1445,9 +1474,18 @@ rule FORTRAN_DEP_HACK
         self.check_outputs(elem)
         return elem
 
+    def determine_rpath_dirs(self, target):
+        link_deps = target.get_all_link_deps()
+        result = []
+        for ld in link_deps:
+            prospective = self.get_target_dir(ld)
+            if not prospective in result:
+                result.append(prospective)
+        return result
+
     def get_dependency_filename(self, t):
         if isinstance(t, build.SharedLibrary):
-            return os.path.join(self.get_target_private_dir(t), self.get_target_filename(t) + '.symbols')
+            return os.path.join(self.get_target_private_dir_abs(t), self.get_target_filename(t) + '.symbols')
         return self.get_target_filename(t)
 
     def generate_shlib_aliases(self, target, outdir):
@@ -1471,7 +1509,7 @@ rule FORTRAN_DEP_HACK
             gcno_elem.add_item('COMMAND', [sys.executable, clean_script, '.', 'gcno'])
             gcno_elem.add_item('description', 'Deleting gcno files')
             gcno_elem.write(outfile)
-            self.check_outputs(elem, gcno_elem)
+            self.check_outputs(gcno_elem, gcno_elem)
 
             gcda_elem = NinjaBuildElement('clean-gcda', 'CUSTOM_COMMAND', 'PHONY')
             script_root = self.environment.get_script_dir()
@@ -1501,7 +1539,7 @@ rule FORTRAN_DEP_HACK
                 plainname = os.path.split(src)[1]
                 basename = plainname.split('.')[0]
                 outname = rule.name_templ.replace('@BASENAME@', basename).replace('@PLAINNAME@', plainname)
-                outfilename = os.path.join(self.get_target_private_dir(target), outname)
+                outfilename = os.path.join(self.get_target_private_dir_abs(target), outname)
                 infilename = os.path.join(self.build_to_src, target.get_source_subdir(), src)
                 elem = NinjaBuildElement(outfilename, rule.name, infilename)
                 elem.write(outfile)
@@ -1529,25 +1567,14 @@ rule FORTRAN_DEP_HACK
         elem = NinjaBuildElement('clean', 'CUSTOM_COMMAND', 'PHONY')
         elem.add_item('COMMAND', [ninja_command, '-t', 'clean'])
         elem.add_item('description', 'Cleaning')
-        if self.environment.coredata.coverage:
+        if self.environment.coredata.get_builtin_option('coverage'):
             self.generate_gcov_clean(outfile)
             elem.add_dep('clean-gcda')
             elem.add_dep('clean-gcno')
         elem.write(outfile)
         self.check_outputs(elem)
 
-        deps = [os.path.join(self.build_to_src, df) \
-                for df in self.interpreter.get_build_def_files()]
-        if self.environment.is_cross_build():
-            deps.append(os.path.join(self.build_to_src,
-                                     self.environment.coredata.cross_file))
-        deps.append('meson-private/coredata.dat')
-        if os.path.exists(os.path.join(self.environment.get_source_dir(), 'meson_options.txt')):
-            deps.append(os.path.join(self.build_to_src, 'meson_options.txt'))
-        for sp in self.build.subprojects.keys():
-            fname = os.path.join(self.environment.get_source_dir(), sp, 'meson_options.txt')
-            if os.path.isfile(fname):
-                deps.append(os.path.join(self.build_to_src, sp, 'meson_options.txt'))
+        deps = self.get_regen_filelist()
         elem = NinjaBuildElement('build.ninja', 'REGENERATE_BUILD', deps)
         elem.add_item('pool', 'console')
         elem.write(outfile)
