@@ -18,15 +18,23 @@ import urllib.request, json
 import sys, os
 import configparser
 import shutil
+import platform
 try:
+    # FIXME for some unknown reason ssl connections
+    # fail on OSX. Thus fall back to unencrypted
+    # traffic.
+    if platform.system().lower() == 'darwin':
+        import nonexisting
     import ssl
     has_ssl = True
     API_ROOT = 'https://wrapdb.mesonbuild.com/v1/'
 except ImportError:
-    print('Warning: ssl not available, traffic not authenticated.')
     has_ssl = False
     API_ROOT = 'http://wrapdb.mesonbuild.com/v1/'
 
+
+ssl_warning_printed = False
+
 from glob import glob
 
 wrapdb_certificate = '''-----BEGIN CERTIFICATE-----
@@ -95,11 +103,23 @@ def build_ssl_context():
     ctx.load_verify_locations(cadata=wrapdb_certificate)
     return ctx
 
-def get_result(urlstring):
+def open_wrapdburl(urlstring):
+    global ssl_warning_printed
     if has_ssl:
-        u = urllib.request.urlopen(urlstring, context=build_ssl_context())
+        return urllib.request.urlopen(urlstring, context=build_ssl_context())
     else:
-        u = urllib.request.urlopen(urlstring)
+        if not ssl_warning_printed:
+            print('Warning: ssl not available, traffic not authenticated.',
+                  file=sys.stderr)
+            ssl_warning_printed = True
+        # Trying to open SSL connection to wrapdb fails because the
+        # certificate is not known. Use plain http. For some
+        # reason OSX fails here again.
+        urlstring = 'http' + urlstring[5:]
+        return urllib.request.urlopen(urlstring)
+
+def get_result(urlstring):
+    u = open_wrapdburl(urlstring)
     data = u.read().decode('utf-8')
     jd = json.loads(data)
     if jd['output'] != 'ok':
@@ -141,7 +161,7 @@ def install(name):
         print('Wrap file already exists.')
         sys.exit(1)
     (branch, revision) = get_latest_version(name)
-    u = urllib.request.urlopen(API_ROOT + 'projects/%s/%s/%s/get_wrap' % (name, branch, revision))
+    u = open_wrapdburl(API_ROOT + 'projects/%s/%s/%s/get_wrap' % (name, branch, revision))
     data = u.read()
     open(wrapfile, 'wb').write(data)
     print('Installed', name, 'branch', branch, 'revision', revision)
@@ -169,7 +189,7 @@ def update(name):
     if new_branch == branch and new_revision == revision:
         print('Project', name, 'is already up to date.')
         sys.exit(0)
-    u = urllib.request.urlopen(API_ROOT + 'projects/%s/%s/%d/get_wrap' % (name, new_branch, new_revision))
+    u = open_wrapdburl(API_ROOT + 'projects/%s/%s/%d/get_wrap' % (name, new_branch, new_revision))
     data = u.read()
     shutil.rmtree(os.path.join('subprojects', subdir), ignore_errors=True)
     try:
